@@ -37,10 +37,11 @@
 #include "misc.h"
 #include "logger.h"
 #include "conffile.h"
+#include "cache.h"
+#include "player.h"
+#include "http.h"
 
-#if LIBAVFORMAT_VERSION_MAJOR >= 53
-# include "avio_evbuffer.h"
-#endif
+#include "avio_evbuffer.h"
 #include "artwork.h"
 
 #ifndef HAVE_LIBEVENT2
@@ -51,24 +52,29 @@
 # include "spotify.h"
 #endif
 
+#if LIBAVCODEC_VERSION_MAJOR <= 53 || (LIBAVCODEC_VERSION_MAJOR == 54 && LIBAVCODEC_VERSION_MINOR <= 34)
+# define AV_CODEC_ID_MJPEG CODEC_ID_MJPEG
+# define AV_CODEC_ID_PNG CODEC_ID_PNG
+# define AV_CODEC_ID_NONE CODEC_ID_NONE
+#endif
+
 static const char *cover_extension[] =
   {
     "jpg", "png",
   };
 
-
 static int
-artwork_read(char *filename, struct evbuffer *evbuf)
+artwork_read(struct evbuffer *evbuf, char *path)
 {
   uint8_t buf[4096];
   struct stat sb;
   int fd;
   int ret;
 
-  fd = open(filename, O_RDONLY);
+  fd = open(path, O_RDONLY);
   if (fd < 0)
     {
-      DPRINTF(E_WARN, L_ART, "Could not open artwork file '%s': %s\n", filename, strerror(errno));
+      DPRINTF(E_WARN, L_ART, "Could not open artwork file '%s': %s\n", path, strerror(errno));
 
       return -1;
     }
@@ -76,7 +82,7 @@ artwork_read(char *filename, struct evbuffer *evbuf)
   ret = fstat(fd, &sb);
   if (ret < 0)
     {
-      DPRINTF(E_WARN, L_ART, "Could not stat() artwork file '%s': %s\n", filename, strerror(errno));
+      DPRINTF(E_WARN, L_ART, "Could not stat() artwork file '%s': %s\n", path, strerror(errno));
 
       goto out_fail;
     }
@@ -104,27 +110,21 @@ artwork_read(char *filename, struct evbuffer *evbuf)
 static int
 rescale_needed(AVCodecContext *src, int max_w, int max_h, int *target_w, int *target_h)
 {
-  int need_rescale;
-
   DPRINTF(E_DBG, L_ART, "Original image dimensions: w %d h %d\n", src->width, src->height);
 
-  need_rescale = 1;
+  *target_w = src->width;
+  *target_h = src->height;
 
-  if ((max_w <= 0) || (max_h <= 0)) /* No valid dimensions, use original */
-    {
-      need_rescale = 0;
+  if ((src->width == 0) || (src->height == 0))         /* Unknown source size, can't rescale */
+    return 0;
 
-      *target_w = src->width;
-      *target_h = src->height;
-    }
-  else if ((src->width <= max_w) && (src->height <= max_h)) /* Smaller than target */
-    {
-      need_rescale = 0;
+  if ((max_w <= 0) || (max_h <= 0))                    /* No valid target dimensions, use original */
+    return 0;
 
-      *target_w = src->width;
-      *target_h = src->height;
-    }
-  else if (src->width * max_h > src->height * max_w)   /* Wider aspect ratio than target */
+  if ((src->width <= max_w) && (src->height <= max_h)) /* Smaller than target */
+    return 0;
+
+  if (src->width * max_h > src->height * max_w)        /* Wider aspect ratio than target */
     {
       *target_w = max_w;
       *target_h = (double)max_w * ((double)src->height / (double)src->width);
@@ -148,11 +148,11 @@ rescale_needed(AVCodecContext *src, int max_w, int max_h, int *target_w, int *ta
 
   DPRINTF(E_DBG, L_ART, "Destination width %d height %d\n", *target_w, *target_h);
 
-  return need_rescale;
+  return 1;
 }
 
 static int
-artwork_rescale(AVFormatContext *src_ctx, int s, int out_w, int out_h, int format, struct evbuffer *evbuf)
+artwork_rescale(struct evbuffer *evbuf, AVFormatContext *src_ctx, int s, int out_w, int out_h)
 {
   uint8_t *buf;
   uint8_t *outbuf;
@@ -181,6 +181,9 @@ artwork_rescale(AVFormatContext *src_ctx, int s, int out_w, int out_h, int forma
 
   src = src_ctx->streams[s]->codec;
 
+  // Avoids threading issue in both ffmpeg and libav that prevents decoding embedded png's
+  src->thread_count = 1;
+
   img_decoder = avcodec_find_decoder(src->codec_id);
   if (!img_decoder)
     {
@@ -202,12 +205,7 @@ artwork_rescale(AVFormatContext *src_ctx, int s, int out_w, int out_h, int forma
     }
 
   /* Set up output */
-#if LIBAVFORMAT_VERSION_MAJOR >= 53 || (LIBAVFORMAT_VERSION_MAJOR == 52 && LIBAVFORMAT_VERSION_MINOR >= 45)
-  /* FFmpeg 0.6 */
   dst_fmt = av_guess_format("image2", NULL, NULL);
-#else
-  dst_fmt = guess_format("image2", NULL, NULL);
-#endif
   if (!dst_fmt)
     {
       DPRINTF(E_LOG, L_ART, "ffmpeg image2 muxer not available\n");
@@ -216,41 +214,19 @@ artwork_rescale(AVFormatContext *src_ctx, int s, int out_w, int out_h, int forma
       goto out_close_src;
     }
 
-#if LIBAVCODEC_VERSION_MAJOR >= 55 || (LIBAVCODEC_VERSION_MAJOR == 54 && LIBAVCODEC_VERSION_MINOR >= 35)
   dst_fmt->video_codec = AV_CODEC_ID_NONE;
 
   /* Try to keep same codec if possible */
-  if ((src->codec_id == AV_CODEC_ID_PNG) && (format & ART_CAN_PNG))
+  if (src->codec_id == AV_CODEC_ID_PNG)
     dst_fmt->video_codec = AV_CODEC_ID_PNG;
-  else if ((src->codec_id == AV_CODEC_ID_MJPEG) && (format & ART_CAN_JPEG))
+  else if (src->codec_id == AV_CODEC_ID_MJPEG)
     dst_fmt->video_codec = AV_CODEC_ID_MJPEG;
 
   /* If not possible, select new codec */
   if (dst_fmt->video_codec == AV_CODEC_ID_NONE)
     {
-      if (format & ART_CAN_PNG)
-	dst_fmt->video_codec = AV_CODEC_ID_PNG;
-      else if (format & ART_CAN_JPEG)
-	dst_fmt->video_codec = AV_CODEC_ID_MJPEG;
+      dst_fmt->video_codec = AV_CODEC_ID_PNG;
     }
-#else
-  dst_fmt->video_codec = CODEC_ID_NONE;
-
-  /* Try to keep same codec if possible */
-  if ((src->codec_id == CODEC_ID_PNG) && (format & ART_CAN_PNG))
-    dst_fmt->video_codec = CODEC_ID_PNG;
-  else if ((src->codec_id == CODEC_ID_MJPEG) && (format & ART_CAN_JPEG))
-    dst_fmt->video_codec = CODEC_ID_MJPEG;
-
-  /* If not possible, select new codec */
-  if (dst_fmt->video_codec == CODEC_ID_NONE)
-    {
-      if (format & ART_CAN_PNG)
-	dst_fmt->video_codec = CODEC_ID_PNG;
-      else if (format & ART_CAN_JPEG)
-	dst_fmt->video_codec = CODEC_ID_MJPEG;
-    }
-#endif
 
   img_encoder = avcodec_find_encoder(dst_fmt->video_codec);
   if (!img_encoder)
@@ -272,18 +248,7 @@ artwork_rescale(AVFormatContext *src_ctx, int s, int out_w, int out_h, int forma
 
   dst_ctx->oformat = dst_fmt;
 
-#if LIBAVFORMAT_VERSION_MAJOR >= 53
   dst_fmt->flags &= ~AVFMT_NOFILE;
-#else
-  ret = snprintf(dst_ctx->filename, sizeof(dst_ctx->filename), "evbuffer:%p", evbuf);
-  if ((ret < 0) || (ret >= sizeof(dst_ctx->filename)))
-    {
-      DPRINTF(E_LOG, L_ART, "Output artwork URL too long\n");
-
-      ret = -1;
-      goto out_free_dst_ctx;
-    }
-#endif
 
 #if LIBAVFORMAT_VERSION_MAJOR >= 54 || (LIBAVFORMAT_VERSION_MAJOR == 53 && LIBAVFORMAT_VERSION_MINOR >= 21)
   dst_st = avformat_new_stream(dst_ctx, NULL);
@@ -317,7 +282,7 @@ artwork_rescale(AVFormatContext *src_ctx, int s, int out_w, int out_h, int forma
 #endif
 
 #if LIBAVCODEC_VERSION_MAJOR >= 55 || (LIBAVCODEC_VERSION_MAJOR == 54 && LIBAVCODEC_VERSION_MINOR >= 35)
-# ifndef FFMPEG_INCOMPATIBLE_API
+# ifndef HAVE_FFMPEG
   dst->pix_fmt = avcodec_find_best_pix_fmt2((enum AVPixelFormat *)img_encoder->pix_fmts, src->pix_fmt, 1, NULL);
 # else
   dst->pix_fmt = avcodec_find_best_pix_fmt_of_list((enum AVPixelFormat *)img_encoder->pix_fmts, src->pix_fmt, 1, NULL);
@@ -462,11 +427,7 @@ artwork_rescale(AVFormatContext *src_ctx, int s, int out_w, int out_h, int forma
   av_free_packet(&pkt);
 
   /* Open output file */
-#if LIBAVFORMAT_VERSION_MAJOR >= 53
   dst_ctx->pb = avio_evbuffer_open(evbuf);
-#else
-  ret = url_fopen(&dst_ctx->pb, dst_ctx->filename, URL_WRONLY);
-#endif
   if (ret < 0)
     {
       DPRINTF(E_LOG, L_ART, "Could not open artwork destination buffer\n");
@@ -485,11 +446,7 @@ artwork_rescale(AVFormatContext *src_ctx, int s, int out_w, int out_h, int forma
     {
       DPRINTF(E_LOG, L_ART, "Out of memory for encoded artwork buffer\n");
 
-#if LIBAVFORMAT_VERSION_MAJOR >= 53
       avio_evbuffer_close(dst_ctx->pb);
-#else
-      url_fclose(dst_ctx->pb);
-#endif
 
       ret = -1;
       goto out_free_buf;
@@ -562,19 +519,11 @@ artwork_rescale(AVFormatContext *src_ctx, int s, int out_w, int out_h, int forma
 
   switch (dst_fmt->video_codec)
     {
-#if LIBAVCODEC_VERSION_MAJOR >= 55 || (LIBAVCODEC_VERSION_MAJOR == 54 && LIBAVCODEC_VERSION_MINOR >= 35)
       case AV_CODEC_ID_PNG:
-#else
-      case CODEC_ID_PNG:
-#endif
 	ret = ART_FMT_PNG;
 	break;
 
-#if LIBAVCODEC_VERSION_MAJOR >= 55 || (LIBAVCODEC_VERSION_MAJOR == 54 && LIBAVCODEC_VERSION_MINOR >= 35)
       case AV_CODEC_ID_MJPEG:
-#else
-      case CODEC_ID_MJPEG:
-#endif
 	ret = ART_FMT_JPEG;
 	break;
 
@@ -585,11 +534,7 @@ artwork_rescale(AVFormatContext *src_ctx, int s, int out_w, int out_h, int forma
     }
 
  out_fclose_dst:
-#if LIBAVFORMAT_VERSION_MAJOR >= 53
   avio_evbuffer_close(dst_ctx->pb);
-#else
-  url_fclose(dst_ctx->pb);
-#endif
   av_free(outbuf);
 
  out_free_buf:
@@ -624,7 +569,7 @@ artwork_rescale(AVFormatContext *src_ctx, int s, int out_w, int out_h, int forma
 }
 
 static int
-artwork_get(char *filename, int max_w, int max_h, int format, struct evbuffer *evbuf)
+artwork_get(struct evbuffer *evbuf, char *path, int max_w, int max_h)
 {
   AVFormatContext *src_ctx;
   int s;
@@ -638,13 +583,13 @@ artwork_get(char *filename, int max_w, int max_h, int format, struct evbuffer *e
   src_ctx = NULL;
 
 #if LIBAVFORMAT_VERSION_MAJOR >= 54 || (LIBAVFORMAT_VERSION_MAJOR == 53 && LIBAVFORMAT_VERSION_MINOR >= 3)
-  ret = avformat_open_input(&src_ctx, filename, NULL, NULL);
+  ret = avformat_open_input(&src_ctx, path, NULL, NULL);
 #else
-  ret = av_open_input_file(&src_ctx, filename, NULL, 0, NULL);
+  ret = av_open_input_file(&src_ctx, path, NULL, 0, NULL);
 #endif
   if (ret < 0)
     {
-      DPRINTF(E_WARN, L_ART, "Cannot open artwork file '%s': %s\n", filename, strerror(AVUNERROR(ret)));
+      DPRINTF(E_WARN, L_ART, "Cannot open artwork file '%s': %s\n", path, strerror(AVUNERROR(ret)));
 
       return -1;
     }
@@ -669,29 +614,21 @@ artwork_get(char *filename, int max_w, int max_h, int format, struct evbuffer *e
   format_ok = 0;
   for (s = 0; s < src_ctx->nb_streams; s++)
     {
-#if LIBAVCODEC_VERSION_MAJOR >= 55 || (LIBAVCODEC_VERSION_MAJOR == 54 && LIBAVCODEC_VERSION_MINOR >= 35)
       if (src_ctx->streams[s]->codec->codec_id == AV_CODEC_ID_PNG)
-#else
-      if (src_ctx->streams[s]->codec->codec_id == CODEC_ID_PNG)
-#endif
 	{
-	  format_ok = (format & ART_CAN_PNG) ? ART_FMT_PNG : 0;
+	  format_ok = ART_FMT_PNG;
 	  break;
 	}
-#if LIBAVCODEC_VERSION_MAJOR >= 55 || (LIBAVCODEC_VERSION_MAJOR == 54 && LIBAVCODEC_VERSION_MINOR >= 35)
       else if (src_ctx->streams[s]->codec->codec_id == AV_CODEC_ID_MJPEG)
-#else
-      else if (src_ctx->streams[s]->codec->codec_id == CODEC_ID_MJPEG)
-#endif
 	{
-	  format_ok = (format & ART_CAN_JPEG) ? ART_FMT_JPEG : 0;
+	  format_ok = ART_FMT_JPEG;
 	  break;
 	}
     }
 
   if (s == src_ctx->nb_streams)
     {
-      DPRINTF(E_LOG, L_ART, "Artwork file '%s' not a PNG or JPEG file\n", filename);
+      DPRINTF(E_LOG, L_ART, "Artwork file '%s' not a PNG or JPEG file\n", path);
 
 #if LIBAVFORMAT_VERSION_MAJOR >= 54 || (LIBAVFORMAT_VERSION_MAJOR == 53 && LIBAVFORMAT_VERSION_MINOR >= 21)
       avformat_close_input(&src_ctx);
@@ -706,12 +643,12 @@ artwork_get(char *filename, int max_w, int max_h, int format, struct evbuffer *e
   /* Fastpath */
   if (!ret && format_ok)
     {
-      ret = artwork_read(filename, evbuf);
+      ret = artwork_read(evbuf, path);
       if (ret == 0)
 	ret = format_ok;
     }
   else
-    ret = artwork_rescale(src_ctx, s, target_w, target_h, format, evbuf);
+    ret = artwork_rescale(evbuf, src_ctx, s, target_w, target_h);
 
 #if LIBAVFORMAT_VERSION_MAJOR >= 54 || (LIBAVFORMAT_VERSION_MAJOR == 53 && LIBAVFORMAT_VERSION_MINOR >= 21)
   avformat_close_input(&src_ctx);
@@ -728,9 +665,87 @@ artwork_get(char *filename, int max_w, int max_h, int format, struct evbuffer *e
   return ret;
 }
 
+/*
+ * Downloads the artwork pointed to by the ICY metadata tag in an internet radio
+ * stream (the StreamUrl tag). The path will be converted back to the id, which
+ * is given to the player. If the id is currently being played, and there is a
+ * valid ICY metadata artwork URL available, it will be returned to this
+ * function, which will then use the http client to get the artwork. Notice: No
+ * rescaling is done.
+ *
+ * @param evbuf the event buffer that will contain the (scaled) image
+ * @param path path to the item we are getting artwork for
+ * @return ART_FMT_* on success, 0 on error and nothing found
+ */
+static int
+artwork_get_player_image(struct evbuffer *evbuf, char *path)
+{
+  struct http_client_ctx ctx;
+  struct keyval *kv;
+  const char *content_type;
+  char *url;
+  int format;
+  int id;
+  int len;
+
+  DPRINTF(E_DBG, L_ART, "Trying internet stream artwork in %s\n", path);
+
+  format = 0;
+
+  id = db_file_id_bypath(path);
+  if (!id)
+    return 0;
+
+  url = player_get_icy_artwork_url(id);
+  if (!url)
+    return 0;
+
+  len = strlen(url);
+  if ((len < 14) || (len > PATH_MAX)) // Can't be shorter than http://a/1.jpg
+    goto out_url;
+
+  cache_artwork_read(evbuf, url, &format);
+  if (format > 0)
+    goto out_url;
+
+  kv = keyval_alloc();
+  if (!kv)
+    goto out_url;
+
+  memset(&ctx, 0, sizeof(ctx));
+  ctx.url = url;
+  ctx.headers = kv;
+  ctx.body = evbuf;
+
+  if (http_client_request(&ctx) < 0)
+    goto out_kv;
+
+  content_type = keyval_get(kv, "Content-Type");
+  if (content_type && (strcmp(content_type, "image/jpeg") == 0))
+    format = ART_FMT_JPEG;
+  else if (content_type && (strcmp(content_type, "image/png") == 0))
+    format = ART_FMT_PNG;
+
+  if (format)
+    {
+      DPRINTF(E_DBG, L_ART, "Found internet stream artwork in %s (%s)\n", url, content_type);
+
+      cache_artwork_stash(evbuf, url, format);
+    }
+
+ out_kv:
+  keyval_clear(kv);
+  free(kv);
+
+ out_url:
+  free(url);
+
+  return format;
+}
+
 #if LIBAVFORMAT_VERSION_MAJOR >= 55 || (LIBAVFORMAT_VERSION_MAJOR == 54 && LIBAVFORMAT_VERSION_MINOR >= 6)
 static int
-artwork_get_embedded_image(char *filename, int max_w, int max_h, int format, struct evbuffer *evbuf)
+artwork_get_embedded_image(struct evbuffer *evbuf, char *path, int max_w, int max_h)
 {
   AVFormatContext *src_ctx;
   AVStream *src_st;
@@ -740,14 +755,14 @@ artwork_get_embedded_image(char *filename, int max_w, int max_h, int format, str
   int format_ok;
   int ret;
 
-  DPRINTF(E_SPAM, L_ART, "Trying embedded artwork in %s\n", filename);
+  DPRINTF(E_SPAM, L_ART, "Trying embedded artwork in %s\n", path);
 
   src_ctx = NULL;
 
-  ret = avformat_open_input(&src_ctx, filename, NULL, NULL);
+  ret = avformat_open_input(&src_ctx, path, NULL, NULL);
   if (ret < 0)
     {
-      DPRINTF(E_WARN, L_ART, "Cannot open media file '%s': %s\n", filename, strerror(AVUNERROR(ret)));
+      DPRINTF(E_WARN, L_ART, "Cannot open media file '%s': %s\n", path, strerror(AVUNERROR(ret)));
 
       return -1;
     }
@@ -766,22 +781,14 @@ artwork_get_embedded_image(char *filename, int max_w, int max_h, int format, str
     {
       if (src_ctx->streams[s]->disposition & AV_DISPOSITION_ATTACHED_PIC)
 	{
-#if LIBAVCODEC_VERSION_MAJOR >= 55 || (LIBAVCODEC_VERSION_MAJOR == 54 && LIBAVCODEC_VERSION_MINOR >= 35)
 	  if (src_ctx->streams[s]->codec->codec_id == AV_CODEC_ID_PNG)
-#else
-	  if (src_ctx->streams[s]->codec->codec_id == CODEC_ID_PNG)
-#endif
 	    {
-	      format_ok = (format & ART_CAN_PNG) ? ART_FMT_PNG : 0;
+	      format_ok = ART_FMT_PNG;
 	      break;
 	    }
-#if LIBAVCODEC_VERSION_MAJOR >= 55 || (LIBAVCODEC_VERSION_MAJOR == 54 && LIBAVCODEC_VERSION_MINOR >= 35)
 	  else if (src_ctx->streams[s]->codec->codec_id == AV_CODEC_ID_MJPEG)
-#else
-	  else if (src_ctx->streams[s]->codec->codec_id == CODEC_ID_MJPEG)
-#endif
 	    {
-	      format_ok = (format & ART_CAN_JPEG) ? ART_FMT_JPEG : 0;
+	      format_ok = ART_FMT_JPEG;
 	      break;
 	    }
 	}
@@ -789,13 +796,13 @@ artwork_get_embedded_image(char *filename, int max_w, int max_h, int format, str
 
   if (s == src_ctx->nb_streams)
     {
-      DPRINTF(E_SPAM, L_ART, "Did not find embedded artwork in '%s'\n", filename);
+      DPRINTF(E_DBG, L_ART, "Did not find embedded artwork in '%s'\n", path);
 
       avformat_close_input(&src_ctx);
       return -1;
     }
   else
-    DPRINTF(E_DBG, L_ART, "Found embedded artwork in '%s'\n", filename);
+    DPRINTF(E_DBG, L_ART, "Found embedded artwork in '%s'\n", path);
 
   src_st = src_ctx->streams[s];
 
@@ -827,7 +834,7 @@ artwork_get_embedded_image(char *filename, int max_w, int max_h, int format, str
     {
       DPRINTF(E_DBG, L_ART, "Artwork too large, rescaling image\n");
 
-      ret = artwork_rescale(src_ctx, s, target_w, target_h, format, evbuf);
+      ret = artwork_rescale(evbuf, src_ctx, s, target_w, target_h);
     }
 
   avformat_close_input(&src_ctx);
@@ -842,280 +849,321 @@ artwork_get_embedded_image(char *filename, int max_w, int max_h, int format, str
 }
 #endif
 
+/*
+ * Looks for basename(in_path).{png,jpg}, so if is in_path is /foo/bar.mp3 it
+ * will look for /foo/bar.png and /foo/bar.jpg
+ *
+ * @param evbuf the event buffer that will contain the (scaled) image
+ * @param in_path path to the item we are getting artwork for
+ * @param max_w maximum image width
+ * @param max_h maximum image height
+ * @param out_path path to artwork, input must be either NULL or char[PATH_MAX]
+ * @return ART_FMT_* on success, 0 on nothing found, -1 on error
+ */
 static int
-artwork_get_own_image(char *path, int max_w, int max_h, int format, struct evbuffer *evbuf)
+artwork_get_own_image(struct evbuffer *evbuf, char *in_path, int max_w, int max_h, char *out_path)
 {
-  char artwork[PATH_MAX];
+  char path[PATH_MAX];
   char *ptr;
   int len;
+  int nextensions;
   int i;
   int ret;
 
-  ret = snprintf(artwork, sizeof(artwork), "%s", path);
-  if ((ret < 0) || (ret >= sizeof(artwork)))
+  ret = snprintf(path, sizeof(path), "%s", in_path);
+  if ((ret < 0) || (ret >= sizeof(path)))
     {
-      DPRINTF(E_INFO, L_ART, "Artwork path exceeds PATH_MAX\n");
-
+      DPRINTF(E_LOG, L_ART, "Artwork path exceeds PATH_MAX (%s)\n", in_path);
       return -1;
     }
 
-  ptr = strrchr(artwork, '.');
+  ptr = strrchr(path, '.');
   if (ptr)
     *ptr = '\0';
 
-  len = strlen(artwork);
+  len = strlen(path);
 
-  for (i = 0; i < (sizeof(cover_extension) / sizeof(cover_extension[0])); i++)
+  nextensions = sizeof(cover_extension) / sizeof(cover_extension[0]);
+
+  for (i = 0; i < nextensions; i++)
     {
-      ret = snprintf(artwork + len, sizeof(artwork) - len, ".%s", cover_extension[i]);
-      if ((ret < 0) || (ret >= sizeof(artwork) - len))
+      ret = snprintf(path + len, sizeof(path) - len, ".%s", cover_extension[i]);
+      if ((ret < 0) || (ret >= sizeof(path) - len))
 	{
-	  DPRINTF(E_INFO, L_ART, "Artwork path exceeds PATH_MAX (ext %s)\n", cover_extension[i]);
-
+	  DPRINTF(E_LOG, L_ART, "Artwork path will exceed PATH_MAX (%s)\n", in_path);
 	  continue;
 	}
 
-      DPRINTF(E_SPAM, L_ART, "Trying own artwork file %s\n", artwork);
+      DPRINTF(E_SPAM, L_ART, "Trying own artwork file %s\n", path);
 
-      ret = access(artwork, F_OK);
+      ret = access(path, F_OK);
       if (ret < 0)
 	continue;
 
       break;
     }
 
-  if (i == (sizeof(cover_extension) / sizeof(cover_extension[0])))
-    return -1;
+  if (i == nextensions)
+    return 0;
 
-  DPRINTF(E_DBG, L_ART, "Found own artwork file %s\n", artwork);
+  DPRINTF(E_DBG, L_ART, "Found own artwork file %s\n", path);
 
-  return artwork_get(artwork, max_w, max_h, format, evbuf);
+  if (out_path)
+    strcpy(out_path, path);
+
+  return artwork_get(evbuf, path, max_w, max_h);
 }
 
+/*
+ * Looks for cover files in a directory, so if dir is /foo/bar and the user has
+ * configured the cover file names "cover" and "artwork" it will look for
+ * /foo/bar/cover.{png,jpg}, /foo/bar/artwork.{png,jpg} and also
+ * /foo/bar/bar.{png,jpg} (so called parentdir artwork)
+ *
+ * @param evbuf the event buffer that will contain the (scaled) image
+ * @param dir the directory to search
+ * @param max_w maximum image width
+ * @param max_h maximum image height
+ * @param out_path path to artwork, input must be either NULL or char[PATH_MAX]
+ * @return ART_FMT_* on success, 0 on nothing found, -1 on error
+ */
 static int
-artwork_get_dir_image(char *path, int isdir, int max_w, int max_h, int format, struct evbuffer *evbuf)
+artwork_get_dir_image(struct evbuffer *evbuf, char *dir, int max_w, int max_h, char *out_path)
 {
-  char artwork[PATH_MAX];
-  char *ptr;
+  char path[PATH_MAX];
+  char parentdir[PATH_MAX];
   int i;
   int j;
   int len;
   int ret;
   cfg_t *lib;
   int nbasenames;
+  int nextensions;
+  char *ptr;
 
-  ret = snprintf(artwork, sizeof(artwork), "%s", path);
-  if ((ret < 0) || (ret >= sizeof(artwork)))
+  ret = snprintf(path, sizeof(path), "%s", dir);
+  if ((ret < 0) || (ret >= sizeof(path)))
     {
-      DPRINTF(E_INFO, L_ART, "Artwork path exceeds PATH_MAX\n");
-
+      DPRINTF(E_LOG, L_ART, "Artwork path exceeds PATH_MAX (%s)\n", dir);
       return -1;
     }
 
-  if (!isdir)
-    {
-      ptr = strrchr(artwork, '/');
-      if (ptr)
-	*ptr = '\0';
-    }
-
-  len = strlen(artwork);
+  len = strlen(path);
 
   lib = cfg_getsec(cfg, "library");
   nbasenames = cfg_size(lib, "artwork_basenames");
 
   if (nbasenames == 0)
-    return -1;
+    return 0;
+
+  nextensions = sizeof(cover_extension) / sizeof(cover_extension[0]);
 
   for (i = 0; i < nbasenames; i++)
     {
-      for (j = 0; j < (sizeof(cover_extension) / sizeof(cover_extension[0])); j++)
+      for (j = 0; j < nextensions; j++)
 	{
-	  ret = snprintf(artwork + len, sizeof(artwork) - len, "/%s.%s", cfg_getnstr(lib, "artwork_basenames", i), cover_extension[j]);
-	  if ((ret < 0) || (ret >= sizeof(artwork) - len))
+	  ret = snprintf(path + len, sizeof(path) - len, "/%s.%s", cfg_getnstr(lib, "artwork_basenames", i), cover_extension[j]);
+	  if ((ret < 0) || (ret >= sizeof(path) - len))
 	    {
-	      DPRINTF(E_INFO, L_ART, "Artwork path exceeds PATH_MAX (%s.%s)\n", cfg_getnstr(lib, "artwork_basenames", i), cover_extension[j]);
-
+	      DPRINTF(E_LOG, L_ART, "Artwork path will exceed PATH_MAX (%s/%s)\n", dir, cfg_getnstr(lib, "artwork_basenames", i));
 	      continue;
 	    }
 
-	  DPRINTF(E_SPAM, L_ART, "Trying directory artwork file %s\n", artwork);
+	  DPRINTF(E_SPAM, L_ART, "Trying directory artwork file %s\n", path);
 
-	  ret = access(artwork, F_OK);
+	  ret = access(path, F_OK);
+	  if (ret < 0)
+	    continue;
+
+	  // If artwork file exists (ret == 0), exit the loop
+	  break;
+	}
+
+      // In case the previous loop exited early, we found an existing artwork file and exit the outer loop
+      if (j < nextensions)
+	break;
+    }
+
+  // If the loop for directory artwork did not exit early, look for parent directory artwork
+  if (i == nbasenames)
+    {
+      ptr = strrchr(path, '/');
+      if (ptr)
+	*ptr = '\0';
+
+      ptr = strrchr(path, '/');
+      if ((!ptr) || (strlen(ptr) <= 1))
+	{
+	  DPRINTF(E_LOG, L_ART, "Could not find parent dir name (%s)\n", path);
+	  return -1;
+	}
+      strcpy(parentdir, ptr + 1);
+
+      len = strlen(path);
+
+      for (i = 0; i < nextensions; i++)
+	{
+	  ret = snprintf(path + len, sizeof(path) - len, "/%s.%s", parentdir, cover_extension[i]);
+	  if ((ret < 0) || (ret >= sizeof(path) - len))
+	    {
+	      DPRINTF(E_LOG, L_ART, "Artwork path will exceed PATH_MAX (%s)\n", parentdir);
+	      continue;
+	    }
+
+	  DPRINTF(E_SPAM, L_ART, "Trying parent directory artwork file %s\n", path);
+
+	  ret = access(path, F_OK);
 	  if (ret < 0)
 	    continue;
 
 	  break;
 	}
 
-      if (j < (sizeof(cover_extension) / sizeof(cover_extension[0])))
-	break;
+      if (i == nextensions)
+	return 0;
     }
 
-  if (i == nbasenames)
-    return -1;
+  DPRINTF(E_DBG, L_ART, "Found directory artwork file %s\n", path);
 
-  DPRINTF(E_DBG, L_ART, "Found directory artwork file %s\n", artwork);
+  if (out_path)
+    strcpy(out_path, path);
 
-  return artwork_get(artwork, max_w, max_h, format, evbuf);
+  return artwork_get(evbuf, path, max_w, max_h);
 }
 
+/*
+ * Given an artwork type (eg embedded, Spotify, own) this function will direct
+ * to the appropriate handler
+ *
+ * @param evbuf the event buffer that will contain the (scaled) image
+ * @param in_path path to the item we are getting artwork for
+ * @param artwork type of the artwork
+ * @param max_w maximum image width
+ * @param max_h maximum image height
+ * @param out_path path to artwork, input must be either NULL or char[PATH_MAX]
+ * @return ART_FMT_* on success, 0 on nothing found, -1 on error
+ */
 static int
-artwork_get_parentdir_image(char *path, int isdir, int max_w, int max_h, int format, struct evbuffer *evbuf)
-{
-  char artwork[PATH_MAX];
-  char parentdir[PATH_MAX];
-  char *ptr;
-  int len;
-  int i;
-  int ret;
-
-  ret = snprintf(artwork, sizeof(artwork), "%s", path);
-  if ((ret < 0) || (ret >= sizeof(artwork)))
-    {
-      DPRINTF(E_INFO, L_ART, "Artwork path exceeds PATH_MAX\n");
-
-      return -1;
-    }
-
-  if (!isdir)
-    {
-      ptr = strrchr(artwork, '/');
-      if (ptr)
-	*ptr = '\0';
-    }
-
-  ptr = strrchr(artwork, '/');
-  if ((!ptr) || (strlen(ptr) <= 1))
-    return -1;
-  strcpy(parentdir, ptr + 1);
-
-  len = strlen(artwork);
-
-  for (i = 0; i < (sizeof(cover_extension) / sizeof(cover_extension[0])); i++)
-    {
-      ret = snprintf(artwork + len, sizeof(artwork) - len, "/%s.%s", parentdir, cover_extension[i]);
-      if ((ret < 0) || (ret >= sizeof(artwork) - len))
-	{
-	  DPRINTF(E_INFO, L_ART, "Artwork path exceeds PATH_MAX (%s.%s)\n", parentdir, cover_extension[i]);
-
-	  continue;
-	}
-
-      DPRINTF(E_SPAM, L_ART, "Trying parent directory artwork file %s\n", artwork);
-
-      ret = access(artwork, F_OK);
-      if (ret < 0)
-	continue;
-
-      break;
-    }
-
-  if (i == (sizeof(cover_extension) / sizeof(cover_extension[0])))
-    return -1;
-
-  DPRINTF(E_DBG, L_ART, "Found parent directory artwork file %s\n", artwork);
-
-  return artwork_get(artwork, max_w, max_h, format, evbuf);
-}
-
-static int
-artwork_get_item_path(char *path, int artwork, uint32_t data_kind, int nodir, int max_w, int max_h, int format, struct evbuffer *evbuf)
+artwork_get_item_path(struct evbuffer *evbuf, char *in_path, int artwork, int max_w, int max_h, char *out_path)
 {
   int ret;
 
   ret = 0;
+  if (out_path)
+    strcpy(out_path, in_path);
+
   switch (artwork)
     {
       case ARTWORK_NONE:
 	break;
+      case ARTWORK_UNKNOWN:
+      case ARTWORK_OWN:
+	if (cfg_getbool(cfg_getsec(cfg, "library"), "artwork_individual"))
+	  ret = artwork_get_own_image(evbuf, in_path, max_w, max_h, out_path);
+	break;
 #ifdef HAVE_SPOTIFY_H
       case ARTWORK_SPOTIFY:
-	if (!(format & ART_CAN_JPEG))
-	  break;
-
-	ret = spotify_artwork_get(evbuf, path, max_w, max_h);
+	ret = spotify_artwork_get(evbuf, in_path, max_w, max_h);
 	(ret < 0) ? (ret = 0) : (ret = ART_FMT_JPEG);
 	break;
 #endif
 #if LIBAVFORMAT_VERSION_MAJOR >= 55 || (LIBAVFORMAT_VERSION_MAJOR == 54 && LIBAVFORMAT_VERSION_MINOR >= 6)
       case ARTWORK_EMBEDDED:
-	ret = artwork_get_embedded_image(path, max_w, max_h, format, evbuf);
-
-	/* Fall through if embedded artwork not found */
-	if (ret > 0)
-	  break;
+	ret = artwork_get_embedded_image(evbuf, in_path, max_w, max_h);
+	break;
 #endif
-      default:
-	/* Not a normal file */
-	if (data_kind != 0)
-	  break;
-
-	/* Look for basename(filename).{png,jpg} */
-	ret = artwork_get_own_image(path, max_w, max_h, format, evbuf);
-	if (ret > 0)
-	  break;
-
-	if (nodir)
-	  break;
-
-	/* Look for basedir(filename)/{artwork,cover}.{png,jpg} */
-	ret = artwork_get_dir_image(path, 0, max_w, max_h, format, evbuf);
-	if (ret > 0)
-	  break;
-
-	/* Look for parentdir(filename).{png,jpg} */
-	ret = artwork_get_parentdir_image(path, 0, max_w, max_h, format, evbuf);
-	if (ret > 0)
-	  break;
+      case ARTWORK_HTTP:
+	ret = artwork_get_player_image(evbuf, in_path);
+	break;
     }
 
-  if (ret > 0)
-    return ret;
-  else
-    return -1;
-}
-
-
-int
-artwork_get_item(int id, int max_w, int max_h, int format, struct evbuffer *evbuf)
-{
-  struct media_file_info *mfi;
-  int ret;
-
-  DPRINTF(E_DBG, L_ART, "Artwork request for item %d\n", id);
-
-  mfi = db_file_fetch_byid(id);
-  if (!mfi)
-    return -1;
-
-  ret = artwork_get_item_path(mfi->path, mfi->artwork, mfi->data_kind, 0, max_w, max_h, format, evbuf);
-  if (ret < 0)
-    DPRINTF(E_DBG, L_ART, "No artwork found for item id %d (%s)\n", id, mfi->fname);
-
-  free_mfi(mfi, 0);
-
   return ret;
 }
 
-int
-artwork_get_group(int id, int max_w, int max_h, int format, struct evbuffer *evbuf)
+/*
+ * Get the artwork for the given media file and the given maxiumum width/height
+
+ * @param evbuf the event buffer that will contain the (scaled) image
+ * @param mfi the media file structure for the file whose image should be returned
+ * @param max_w maximum image width
+ * @param max_h maximum image height
+ * @return ART_FMT_* on success, 0 on nothing found, -1 on error
+ */
+static int
+artwork_get_item_mfi(struct evbuffer *evbuf, struct media_file_info *mfi, int max_w, int max_h)
+{
+  char path[PATH_MAX];
+  int cached;
+  int format;
+  int ret;
+
+  DPRINTF(E_DBG, L_ART, "Looking for artwork for item with id %d\n", mfi->id);
+
+  ret = cache_artwork_get(CACHE_ARTWORK_INDIVIDUAL, mfi->id, max_w, max_h, &cached, &format, evbuf);
+  if ((ret == 0) && cached)
+    {
+      DPRINTF(E_DBG, L_ART, "Item %d found in cache with format %d\n", mfi->id, format);
+      return format;
+    }
+
+  if (mfi->data_kind == DATA_KIND_FILE)
+    {
+      format = artwork_get_item_path(evbuf, mfi->path, mfi->artwork, max_w, max_h, path);
+      
+      if (format > 0)
+	cache_artwork_add(CACHE_ARTWORK_INDIVIDUAL, mfi->id, max_w, max_h, format, path, evbuf);
+
+      return format;
+    }
+
+  return -1;
+}
+
+/*
+ * Get the artwork image for the given persistentid and the given maximum width/height
+ *
+ * The function first checks if there is a cache entry, if not it will first look for directory artwork files.
+ * If no directory artwork files are found, it looks for individual artwork (embedded images or images from spotify).
+ *
+ * @param evbuf the event buffer that will contain the (scaled) image
+ * @param persistentid persistent songalbumid or songartistid
+ * @param max_w maximum image width
+ * @param max_h maximum image height
+ * @return ART_FMT_* on success, 0 on nothing found, -1 on error
+ */
+static int
+artwork_get_group_persistentid(struct evbuffer *evbuf, int64_t persistentid, int max_w, int max_h)
 {
   struct query_params qp;
   struct db_media_file_info dbmfi;
+  char path[PATH_MAX];
   char *dir;
-  int got_art;
+  int cached;
+  int format;
   int ret;
   int artwork;
+  int got_spotifyitem;
   uint32_t data_kind;
 
-  DPRINTF(E_DBG, L_ART, "Artwork request for group %d\n", id);
+  DPRINTF(E_DBG, L_ART, "Looking for artwork for group with persistentid %" PRIi64 "\n", persistentid);
 
-  /* Try directory artwork first */
+  got_spotifyitem = 0;
+
+  /*
+   * First check if the artwork cache has a cached entry for the given persistent id and requested width/height
+   */
+  ret = cache_artwork_get(CACHE_ARTWORK_GROUP, persistentid, max_w, max_h, &cached, &format, evbuf);
+  if ((ret == 0) && cached)
+    {
+      DPRINTF(E_DBG, L_ART, "Group %" PRIi64 " found in cache with format %d\n", persistentid, format);
+      return format;
+    }
+
+  /* Image is not in the artwork cache. Try directory artwork first */
   memset(&qp, 0, sizeof(struct query_params));
 
   qp.type = Q_GROUP_DIRS;
-  qp.id = id;
+  qp.persistentid = persistentid;
 
   ret = db_query_start(&qp);
   if (ret < 0)
@@ -1126,27 +1174,33 @@ artwork_get_group(int id, int max_w, int max_h, int format, struct evbuffer *evb
       goto files_art;
     }
 
-  got_art = 0;
-  while (!got_art && ((ret = db_query_fetch_string(&qp, &dir)) == 0) && (dir))
+  format = 0;
+  while (((ret = db_query_fetch_string(&qp, &dir)) == 0) && (dir))
     {
-      /* If item is an internet stream don't look for artwork */
-      if (strncmp(dir, "http://", strlen("http://")) == 0)
+      /* The db query may return non-directories (eg if item is an internet stream or Spotify) */
+      if (access(dir, F_OK) < 0)
 	continue;
 
-      /* If Spotify item don't look for artwork */
-      if (strncmp(dir, "spotify:", strlen("spotify:")) == 0)
-	continue;
+      format = artwork_get_dir_image(evbuf, dir, max_w, max_h, path);
 
-      got_art = (artwork_get_dir_image(dir, 1, max_w, max_h, format, evbuf) > 0)
-                || (artwork_get_parentdir_image(dir, 1, max_w, max_h, format, evbuf) > 0);
+      if (format > 0)
+	break;	
     }
 
   db_query_end(&qp);
 
   if (ret < 0)
-    DPRINTF(E_LOG, L_ART, "Error fetching Q_GROUP_DIRS results\n");
-  else if (got_art > 0)
-    return got_art;
+    { 
+      DPRINTF(E_LOG, L_ART, "Error fetching Q_GROUP_DIRS results\n");
+      goto files_art;
+    }
+
+  /* Found artwork, cache it and return */  
+  if (format > 0)
+    {
+      cache_artwork_add(CACHE_ARTWORK_GROUP, persistentid, max_w, max_h, format, path, evbuf);
+      return format;
+    }
 
 
   /* Then try individual files */
@@ -1154,33 +1208,170 @@ artwork_get_group(int id, int max_w, int max_h, int format, struct evbuffer *evb
   memset(&qp, 0, sizeof(struct query_params));
 
   qp.type = Q_GROUP_ITEMS;
-  qp.id = id;
+  qp.persistentid = persistentid;
 
   ret = db_query_start(&qp);
   if (ret < 0)
     {
       DPRINTF(E_LOG, L_ART, "Could not start Q_GROUP_ITEMS query\n");
-
       return -1;
     }
 
-  got_art = 0;
-  while (!got_art && ((ret = db_query_fetch_file(&qp, &dbmfi)) == 0) && (dbmfi.id))
+  format = 0;
+  while (((ret = db_query_fetch_file(&qp, &dbmfi)) == 0) && (dbmfi.id))
     {
       if ((safe_atoi32(dbmfi.artwork, &artwork) != 0) && (safe_atou32(dbmfi.data_kind, &data_kind) != 0))
 	continue;
 
-      got_art = (artwork_get_item_path(dbmfi.path, artwork, data_kind, 1, max_w, max_h, format, evbuf) > 0);
+      format = artwork_get_item_path(evbuf, dbmfi.path, artwork, max_w, max_h, path);
+
+      if (artwork == ARTWORK_SPOTIFY)
+	got_spotifyitem = 1;
+
+      if (format > 0)
+	break;
     }
 
   db_query_end(&qp);
 
   if (ret < 0)
-    DPRINTF(E_LOG, L_ART, "Error fetching Q_GROUP_ITEMS results\n");
-  else if (got_art > 0)
-    return got_art;
+    {
+      DPRINTF(E_LOG, L_ART, "Error fetching Q_GROUP_ITEMS results\n");
+      return -1;
+    }
 
-  DPRINTF(E_DBG, L_ART, "No artwork found for group %d\n", id);
+  /* Found artwork, cache it and return */
+  if (format > 0)
+    {
+      if (artwork != ARTWORK_HTTP)
+	cache_artwork_add(CACHE_ARTWORK_GROUP, persistentid, max_w, max_h, format, path, evbuf);
+      return format;
+    }
+  else if (format < 0)
+    {
+      DPRINTF(E_WARN, L_ART, "Error getting artwork for group %" PRIi64 "\n", persistentid);
+      return -1;
+    }
 
-  return -1;
+  DPRINTF(E_DBG, L_ART, "No artwork found for group %" PRIi64 "\n", persistentid);
+
+  /* Add cache entry for no artwork available */
+  if ((artwork != ARTWORK_HTTP) && (!got_spotifyitem))
+    cache_artwork_add(CACHE_ARTWORK_GROUP, persistentid, max_w, max_h, 0, "", evbuf);
+
+  return 0;
+}
+
+/*
+ * Get the artwork image for the given item id and the given maximum width/height
+ *
+ * @param evbuf the event buffer that will contain the (scaled) image
+ * @param id the mfi item id
+ * @param max_w maximum image width
+ * @param max_h maximum image height
+ * @return ART_FMT_* on success, -1 on error or no artwork found
+ */
+int
+artwork_get_item(struct evbuffer *evbuf, int id, int max_w, int max_h)
+{
+  struct media_file_info *mfi;
+  int format;
+
+  mfi = db_file_fetch_byid(id);
+  if (!mfi)
+    {
+      DPRINTF(E_LOG, L_ART, "Artwork request for item %d, but no such item in database\n", id);
+      return -1;
+    }
+
+  DPRINTF(E_DBG, L_ART, "Artwork request for item %d (%s)\n", id, mfi->fname);
+
+  format = 0;
+  if (cfg_getbool(cfg_getsec(cfg, "library"), "artwork_individual"))
+    format = artwork_get_item_mfi(evbuf, mfi, max_w, max_h);
+
+  /* No individual artwork or individual artwork disabled, try group artwork */
+  if (format <= 0)
+    format = artwork_get_group_persistentid(evbuf, mfi->songalbumid, max_w, max_h);
+
+  free_mfi(mfi, 0);
+
+  if (format <= 0)
+    {
+      DPRINTF(E_DBG, L_ART, "No artwork found for item %d\n", id);
+      return -1;
+    }
+
+  return format;
+}
+
+/*
+ * Get the artwork image for the given group id and the given maximum width/height
+ *
+ * @param evbuf the event buffer that will contain the (scaled) image
+ * @param id the group id (not the persistent id)
+ * @param max_w maximum image width
+ * @param max_h maximum image height
+ * @return ART_FMT_* on success, -1 on error or no artwork found
+ */
+int
+artwork_get_group(struct evbuffer *evbuf, int id, int max_w, int max_h)
+{
+  int64_t persistentid;
+  int format;
+
+  DPRINTF(E_DBG, L_ART, "Artwork request for group %d\n", id);
+
+  /* Get the persistent id for the given group id */
+  if (db_group_persistentid_byid(id, &persistentid) < 0)
+    {
+      DPRINTF(E_LOG, L_ART, "Error fetching persistent id for group id %d\n", id);
+      return -1;
+    }
+
+  /* Load artwork image for the persistent id */
+  format = artwork_get_group_persistentid(evbuf, persistentid, max_w, max_h);
+  if (format <= 0)
+    {
+      DPRINTF(E_DBG, L_ART, "No artwork found for group %d\n", id);
+      return -1;
+    }
+
+  return format;
+}
+
+/* Checks if the file is an artwork file */
+int
+artwork_file_is_artwork(const char *filename)
+{
+  cfg_t *lib;
+  int n;
+  int i;
+  int j;
+  int ret;
+  char artwork[PATH_MAX];
+
+  lib = cfg_getsec(cfg, "library");
+  n = cfg_size(lib, "artwork_basenames");
+
+  for (i = 0; i < n; i++)
+    {
+      for (j = 0; j < (sizeof(cover_extension) / sizeof(cover_extension[0])); j++)
+	{
+	  ret = snprintf(artwork, sizeof(artwork), "%s.%s", cfg_getnstr(lib, "artwork_basenames", i), cover_extension[j]);
+	  if ((ret < 0) || (ret >= sizeof(artwork)))
+	    {
+	      DPRINTF(E_INFO, L_ART, "Artwork path exceeds PATH_MAX (%s.%s)\n", cfg_getnstr(lib, "artwork_basenames", i), cover_extension[j]);
+	      continue;
+	    }
+
+	  if (strcmp(artwork, filename) == 0)
+	    return 1;
+	}
+
+      if (j < (sizeof(cover_extension) / sizeof(cover_extension[0])))
+	break;
+    }
+
+  return 0;
 }
