@@ -22,22 +22,33 @@ enum sort_type {
   S_ALBUM,
   S_ARTIST,
   S_PLAYLIST,
+  S_YEAR,
+  S_GENRE,
+  S_COMPOSER,
+  S_DISC,
+  S_TRACK,
+  S_VPATH,
 };
 
 #define Q_F_BROWSE (1 << 15)
 
 enum query_type {
-  Q_ITEMS            = (1 << 0),
-  Q_PL               = (1 << 1),
-  Q_PLITEMS          = (1 << 2),
-  Q_BROWSE_ARTISTS   = Q_F_BROWSE | (1 << 3),
-  Q_BROWSE_ALBUMS    = Q_F_BROWSE | (1 << 4),
-  Q_BROWSE_GENRES    = Q_F_BROWSE | (1 << 5),
-  Q_BROWSE_COMPOSERS = Q_F_BROWSE | (1 << 6),
-  Q_GROUP_ALBUMS     = (1 << 7),
-  Q_GROUP_ARTISTS    = (1 << 8),
-  Q_GROUP_ITEMS      = (1 << 9),
-  Q_GROUP_DIRS       = Q_F_BROWSE | (1 << 10),
+  Q_ITEMS            = 1,
+  Q_PL               = 2,
+  Q_PLITEMS          = 3,
+  Q_BROWSE_ARTISTS   = Q_F_BROWSE | 4,
+  Q_BROWSE_ALBUMS    = Q_F_BROWSE | 5,
+  Q_BROWSE_GENRES    = Q_F_BROWSE | 6,
+  Q_BROWSE_COMPOSERS = Q_F_BROWSE | 7,
+  Q_GROUP_ALBUMS     = 8,
+  Q_GROUP_ARTISTS    = 9,
+  Q_GROUP_ITEMS      = 10,
+  Q_GROUP_DIRS       = Q_F_BROWSE | 11,
+  Q_BROWSE_YEARS     = Q_F_BROWSE | 12,
+  Q_COUNT_ITEMS      = 13,
+  Q_BROWSE_DISCS     = Q_F_BROWSE | 14,
+  Q_BROWSE_TRACKS    = Q_F_BROWSE | 15,
+  Q_BROWSE_VPATH     = Q_F_BROWSE | 16,
 };
 
 #define ARTWORK_UNKNOWN   0
@@ -47,6 +58,13 @@ enum query_type {
 #define ARTWORK_DIR       4
 #define ARTWORK_PARENTDIR 5
 #define ARTWORK_SPOTIFY   6
+#define ARTWORK_HTTP      7
+
+enum filelistitem_type {
+  F_PLAYLIST = 1,
+  F_DIR  = 2,
+  F_FILE = 3,
+};
 
 struct query_params {
   /* Query parameters, filled in by caller */
@@ -54,6 +72,7 @@ struct query_params {
   enum index_type idx_type;
   enum sort_type sort;
   int id;
+  int64_t persistentid;
   int offset;
   int limit;
 
@@ -64,7 +83,8 @@ struct query_params {
 
   /* Private query context, keep out */
   sqlite3_stmt *stmt;
-  char buf[32];
+  char buf1[32];
+  char buf2[32];
 };
 
 struct pairing_info {
@@ -73,6 +93,23 @@ struct pairing_info {
   char *guid;
 };
 
+enum media_kind {
+  MEDIA_KIND_MUSIC = 1,
+  MEDIA_KIND_MOVIE = 2,
+  MEDIA_KIND_PODCAST = 4,
+  MEDIA_KIND_AUDIOBOOK = 8,
+  MEDIA_KIND_MUSICVIDEO = 32,
+  MEDIA_KIND_TVSHOW = 64,
+};
+
+enum data_kind {
+  DATA_KIND_FILE = 0,    /* normal file */
+  DATA_KIND_URL = 1,     /* url/stream */
+  DATA_KIND_SPOTIFY = 2, /* iTunes has no spotify data kind, but we use 2 */
+  DATA_KIND_PIPE = 3,    /* iTunes has no pipe data kind, but we use 3 */
+};
+
+/* Note that fields marked as integers in the metadata map in filescanner_ffmpeg must be uint32_t here */
 struct media_file_info {
   char *path;
   uint32_t index;
@@ -121,7 +158,7 @@ struct media_file_info {
   uint32_t item_kind;    /* song or movie */
   uint32_t data_kind;    /* dmap.datakind (asdk) */
   uint64_t sample_count;
-  char compilation;
+  uint32_t compilation;
   char artwork;
 
   /* iTunes 5+ */
@@ -148,14 +185,19 @@ struct media_file_info {
   char *album_sort;
   char *composer_sort;
   char *album_artist_sort;
+
+  char *virtual_path;
 };
 
 #define mfi_offsetof(field) offsetof(struct media_file_info, field)
 
+/* PL_SPECIAL value must be in sync with type value in Q_PL* in db.c */
 enum pl_type {
-  PL_PLAIN,
-  PL_SMART,
-  PL_MAX
+  PL_SPECIAL = 0,
+  PL_FOLDER = 1,
+  PL_SMART = 2,
+  PL_PLAIN = 3,
+  PL_MAX,
 };
 
 struct playlist_info {
@@ -163,12 +205,15 @@ struct playlist_info {
   char *title;           /* playlist name as displayed in iTunes (minm) */
   enum pl_type type;     /* see PL_ types */
   uint32_t items;        /* number of items (mimc) */
+  uint32_t streams;      /* number of internet streams */
   char *query;           /* where clause if type 1 (MSPS) */
   uint32_t db_timestamp; /* time last updated */
   uint32_t disabled;
   char *path;            /* path of underlying playlist */
   uint32_t index;        /* index of playlist for paths with multiple playlists */
   uint32_t special_id;   /* iTunes identifies certain 'special' playlists with special meaning */
+  char *virtual_path;    /* virtual path of underlying playlist */
+  uint32_t parent_id;    /* Id of parent playlist if the playlist is nested */
 };
 
 #define pli_offsetof(field) offsetof(struct playlist_info, field)
@@ -178,12 +223,15 @@ struct db_playlist_info {
   char *title;
   char *type;
   char *items;
+  char *streams;
   char *query;
   char *db_timestamp;
   char *disabled;
   char *path;
   char *index;
   char *special_id;
+  char *virtual_path;
+  char *parent_id;
 };
 
 #define dbpli_offsetof(field) offsetof(struct db_playlist_info, field)
@@ -272,10 +320,17 @@ struct db_media_file_info {
   char *album_sort;
   char *composer_sort;
   char *album_artist_sort;
+  char *virtual_path;
 };
 
 #define dbmfi_offsetof(field) offsetof(struct db_media_file_info, field)
 
+struct filelist_info {
+  char *virtual_path;
+  uint32_t time_modified;
+  enum filelistitem_type type;
+};
+
 struct watch_info {
   int wd;
   char *path;
@@ -292,6 +347,10 @@ struct watch_enum {
   sqlite3_stmt *stmt;
 };
 
+struct filecount_info {
+  uint32_t count;
+  uint32_t length;
+};
 
 char *
 db_escape_string(const char *str);
@@ -299,6 +358,9 @@ db_escape_string(const char *str);
 void
 free_pi(struct pairing_info *pi, int content_only);
 
+void
+free_fi(struct filelist_info *fi, int content_only);
+
 void
 free_mfi(struct media_file_info *mfi, int content_only);
 
@@ -341,6 +403,9 @@ db_query_fetch_pl(struct query_params *qp, struct db_playlist_info *dbpli);
 int
 db_query_fetch_group(struct query_params *qp, struct db_group_info *dbgri);
 
+int
+db_query_fetch_count(struct query_params *qp, struct filecount_info *fci);
+
 int
 db_query_fetch_string(struct query_params *qp, char **string);
 
@@ -393,12 +458,18 @@ db_file_stamp_bypath(char *path, time_t *stamp, int *id);
 struct media_file_info *
 db_file_fetch_byid(int id);
 
+struct media_file_info *
+db_file_fetch_byvirtualpath(char *path);
+
 int
 db_file_add(struct media_file_info *mfi);
 
 int
 db_file_update(struct media_file_info *mfi);
 
+void
+db_file_update_icy(int id, char *artist, char *album);
+
 void
 db_file_delete_bypath(char *path);
 
@@ -424,11 +495,14 @@ db_pl_ping_bymatch(char *path, int isdir);
 struct playlist_info *
 db_pl_fetch_bypath(char *path);
 
+struct playlist_info *
+db_pl_fetch_byvirtualpath(char *virtual_path);
+
 struct playlist_info *
 db_pl_fetch_bytitlepath(char *title, char *path);
 
 int
-db_pl_add(char *title, char *path, int *id);
+db_pl_add(struct playlist_info *pli, int *id);
 
 int
 db_pl_add_item_bypath(int plid, char *path);
@@ -440,7 +514,7 @@ void
 db_pl_clear_items(int id);
 
 int
-db_pl_update(char *title, char *path, int id);
+db_pl_update(struct playlist_info *pli);
 
 void
 db_pl_delete(int id);
@@ -461,8 +535,15 @@ db_pl_enable_bycookie(uint32_t cookie, char *path);
 int
 db_groups_clear(void);
 
-enum group_type
-db_group_type_byid(int id);
+int
+db_group_persistentid_byid(int id, int64_t *persistentid);
+
+/* Filelist */
+int
+db_mpd_start_query_filelist(struct query_params *qp, char *path);
+
+int
+db_mpd_query_fetch_filelist(struct query_params *qp, struct filelist_info *fi);
 
 /* Remotes */
 int
