@@ -53,6 +53,10 @@
 # include <sys/eventfd.h>
 #endif
 
+#ifdef HAVE_REGEX_H
+# include <regex.h>
+#endif
+
 #include <event.h>
 
 #include "logger.h"
@@ -62,6 +66,8 @@
 #include "misc.h"
 #include "remote_pairing.h"
 #include "player.h"
+#include "cache.h"
+#include "artwork.h"
 
 #ifdef LASTFM
 # include "lastfm.h"
@@ -71,6 +77,22 @@
 #endif
 
 
+struct filescanner_command;
+
+typedef int (*cmd_func)(struct filescanner_command *cmd);
+
+struct filescanner_command
+{
+  pthread_mutex_t lck;
+  pthread_cond_t cond;
+
+  cmd_func func;
+
+  int nonblock;
+
+  int ret;
+};
+
 #define F_SCAN_BULK    (1 << 0)
 #define F_SCAN_RESCAN  (1 << 1)
 #define F_SCAN_FAST    (1 << 2)
@@ -81,6 +103,7 @@ enum file_type {
   FILE_IGNORE,
   FILE_REGULAR,
   FILE_PLAYLIST,
+  FILE_SMARTPL,
   FILE_ITUNES,
   FILE_ARTWORK,
   FILE_CTRL_REMOTE,
@@ -102,6 +125,7 @@ struct stacked_dir {
 };
 
 
+static int cmd_pipe[2];
 #ifdef USE_EVENTFD
 static int exit_efd;
 #else
@@ -112,6 +136,7 @@ static int inofd;
 static struct event_base *evbase_scan;
 static struct event inoev;
 static struct event exitev;
+static struct event cmdev;
 static pthread_t tid_scan;
 static struct deferred_pl *playlists;
 static struct stacked_dir *dirstack;
@@ -119,6 +144,19 @@ static struct stacked_dir *dirstack;
 /* Count of files scanned during a bulk scan */
 static int counter;
 
+/* Flag for scan in progress */
+static int scanning;
+
+/* When copying into the lib (eg. if a file is moved to the lib by copying into
+ * a Samba network share) inotify might give us IN_CREATE -> n x IN_ATTRIB ->
+ * IN_CLOSE_WRITE, but we don't want to do any scanning before the
+ * IN_CLOSE_WRITE. So we register new files (by path hashes) in this ring buffer
+ * when we get the IN_CREATE and then ignore the IN_ATTRIB for these files.
+ */
+#define INCOMINGFILES_BUFFER_SIZE 50
+int incomingfiles_idx;
+uint32_t incomingfiles_buffer[INCOMINGFILES_BUFFER_SIZE];
+
 /* Forward */
 static void
 bulk_scan(int flags);
@@ -126,6 +164,46 @@ static int
 inofd_event_set(void);
 static void
 inofd_event_unset(void);
+static int
+filescanner_initscan(struct filescanner_command *cmd);
+static int
+filescanner_fullrescan(struct filescanner_command *cmd);
+
+
+/* ---------------------------- COMMAND EXECUTION -------------------------- */
+
+static int
+send_command(struct filescanner_command *cmd)
+{
+  int ret;
+
+  if (!cmd->func)
+    {
+      DPRINTF(E_LOG, L_SCAN, "BUG: cmd->func is NULL!\n");
+      return -1;
+    }
+
+  ret = write(cmd_pipe[1], &cmd, sizeof(cmd));
+  if (ret != sizeof(cmd))
+    {
+      DPRINTF(E_LOG, L_SCAN, "Could not send command: %s\n", strerror(errno));
+      return -1;
+    }
+
+  return 0;
+}
+
+static int
+nonblock_command(struct filescanner_command *cmd)
+{
+  int ret;
+
+  ret = send_command(cmd);
+  if (ret < 0)
+    return -1;
+
+  return 0;
+}
 
 static int
 push_dir(struct stacked_dir **s, char *path)
@@ -170,6 +248,43 @@ pop_dir(struct stacked_dir **s)
   return ret;
 }
 
+#ifdef HAVE_REGEX_H
+/* Checks if the file path is configured to be ignored */
+static int
+file_path_ignore(const char *path)
+{
+  cfg_t *lib;
+  regex_t regex;
+  int n;
+  int i;
+  int ret;
+
+  lib = cfg_getsec(cfg, "library");
+  n = cfg_size(lib, "filepath_ignore");
+
+  for (i = 0; i < n; i++)
+    {
+      ret = regcomp(&regex, cfg_getnstr(lib, "filepath_ignore", i), 0);
+      if (ret != 0)
+	{
+	  DPRINTF(E_LOG, L_SCAN, "Could not compile regex for matching with file path\n");
+	  return 0;
+	}
+
+      ret = regexec(&regex, path, 0, NULL, 0);
+      regfree(&regex);
+
+      if (ret == 0)
+	{
+	  DPRINTF(E_DBG, L_SCAN, "Regex match: %s\n", path);
+	  return 1;
+	}
+    }
+
+  return 0;
+}
+#endif
+
 /* Checks if the file extension is in the ignore list */
 static int
 file_type_ignore(const char *ext)
@@ -201,6 +316,11 @@ file_type_get(const char *path) {
   else
     filename++;
 
+#ifdef HAVE_REGEX_H
+  if (file_path_ignore(path))
+    return FILE_IGNORE;
+#endif
+
   ext = strrchr(path, '.');
   if (!ext || (strlen(ext) == 1))
     return FILE_REGULAR;
@@ -208,9 +328,15 @@ file_type_get(const char *path) {
   if ((strcasecmp(ext, ".m3u") == 0) || (strcasecmp(ext, ".pls") == 0))
     return FILE_PLAYLIST;
 
-  if ((strcasecmp(ext, ".png") == 0) || (strcasecmp(ext, ".jpg") == 0))
+  if (strcasecmp(ext, ".smartpl") == 0)
+    return FILE_SMARTPL;
+
+  if (artwork_file_is_artwork(filename))
     return FILE_ARTWORK;
 
+  if ((strcasecmp(ext, ".jpg") == 0) || (strcasecmp(ext, ".png") == 0))
+    return FILE_IGNORE;
+
 #ifdef ITUNES
   if (strcasecmp(ext, ".xml") == 0)
     return FILE_ITUNES;
@@ -421,7 +547,7 @@ fixup_tags(struct media_file_info *mfi)
   /* Handle TV shows, try to present prettier metadata */
   if (mfi->tv_series_name && strlen(mfi->tv_series_name) != 0)
     {
-      mfi->media_kind = 64;  /* tv show */
+      mfi->media_kind = MEDIA_KIND_TVSHOW;  /* tv show */
 
       /* Default to artist = series_name */
       if (mfi->artist && strlen(mfi->artist) == 0)
@@ -462,7 +588,7 @@ fixup_tags(struct media_file_info *mfi)
       /* fname is left untouched by unicode_fixup_mfi() for
        * obvious reasons, so ensure it is proper UTF-8
        */
-      mfi->title = unicode_fixup_string(mfi->fname);
+      mfi->title = unicode_fixup_string(mfi->fname, "ascii");
       if (mfi->title == mfi->fname)
 	mfi->title = strdup(mfi->fname);
     }
@@ -492,7 +618,7 @@ fixup_tags(struct media_file_info *mfi)
 	  mfi->album_artist_sort = strdup("");
 	}
     }
-  else if (mfi->media_kind == 4) /* Podcast */
+  else if (mfi->media_kind == MEDIA_KIND_PODCAST) /* Podcast */
     {
       if (mfi->album_artist)
 	free(mfi->album_artist);
@@ -522,6 +648,7 @@ filescanner_process_media(char *path, time_t mtime, off_t size, int type, struct
   char *filename;
   time_t stamp;
   int id;
+  char virtual_path[PATH_MAX];
   int ret;
 
   filename = strrchr(path, '/');
@@ -574,26 +701,30 @@ filescanner_process_media(char *path, time_t mtime, off_t size, int type, struct
 
   if (type & F_SCAN_TYPE_FILE)
     {
-      mfi->data_kind = 0; /* real file */
+      mfi->data_kind = DATA_KIND_FILE;
       ret = scan_metadata_ffmpeg(path, mfi);
     }
   else if (type & F_SCAN_TYPE_URL)
     {
-      mfi->data_kind = 1; /* url/stream */
-#if LIBAVFORMAT_VERSION_MAJOR >= 56 || (LIBAVFORMAT_VERSION_MAJOR == 55 && LIBAVFORMAT_VERSION_MINOR >= 13)
+      mfi->data_kind = DATA_KIND_URL;
       ret = scan_metadata_ffmpeg(path, mfi);
-#else
-      ret = scan_metadata_icy(path, mfi);
-#endif
+      if (ret < 0)
+	{
+	  DPRINTF(E_LOG, L_SCAN, "Playlist URL is unavailable for probe/metadata, assuming MP3 encoding\n");
+	  mfi->type = strdup("mp3");
+	  mfi->codectype = strdup("mpeg");
+	  mfi->description = strdup("MPEG audio file");
+	  ret = 1;
+	}
     }
   else if (type & F_SCAN_TYPE_SPOTIFY)
     {
-      mfi->data_kind = 2; /* iTunes has no spotify data kind, but we use 2 */
+      mfi->data_kind = DATA_KIND_SPOTIFY;
       ret = mfi->artist && mfi->album && mfi->title;
     }
   else if (type & F_SCAN_TYPE_PIPE)
     {
-      mfi->data_kind = 3; /* iTunes has no pipe data kind, but we use 3 */
+      mfi->data_kind = DATA_KIND_PIPE;
       mfi->type = strdup("wav");
       mfi->codectype = strdup("wav");
       mfi->description = strdup("PCM16 pipe");
@@ -614,19 +745,35 @@ filescanner_process_media(char *path, time_t mtime, off_t size, int type, struct
   if (type & F_SCAN_TYPE_COMPILATION)
     mfi->compilation = 1;
   if (type & F_SCAN_TYPE_PODCAST)
-    mfi->media_kind = 4; /* podcast */
+    mfi->media_kind = MEDIA_KIND_PODCAST; /* podcast */
   if (type & F_SCAN_TYPE_AUDIOBOOK)
-    mfi->media_kind = 8; /* audiobook */
+    mfi->media_kind = MEDIA_KIND_AUDIOBOOK; /* audiobook */
 
   if (!mfi->item_kind)
     mfi->item_kind = 2; /* music */
   if (!mfi->media_kind)
-    mfi->media_kind = 1; /* music */
+    mfi->media_kind = MEDIA_KIND_MUSIC; /* music */
 
   unicode_fixup_mfi(mfi);
 
   fixup_tags(mfi);
 
+  if (type & F_SCAN_TYPE_URL)
+    {
+      snprintf(virtual_path, PATH_MAX, "/http:/%s", mfi->title);
+      mfi->virtual_path = strdup(virtual_path);
+    }
+  else if (type & F_SCAN_TYPE_SPOTIFY)
+    {
+      snprintf(virtual_path, PATH_MAX, "/spotify:/%s/%s/%s", mfi->album_artist, mfi->album, mfi->title);
+      mfi->virtual_path = strdup(virtual_path);
+    }
+  else
+    {
+      snprintf(virtual_path, PATH_MAX, "/file:%s", mfi->path);
+      mfi->virtual_path = strdup(virtual_path);
+    }
+
   if (mfi->id == 0)
     db_file_add(mfi);
   else
@@ -707,11 +854,18 @@ process_deferred_playlists(void)
 static void
 process_file(char *file, time_t mtime, off_t size, int type, int flags)
 {
+  int is_bulkscan;
+
+  is_bulkscan = (flags & F_SCAN_BULK);
+
   switch (file_type_get(file))
     {
       case FILE_REGULAR:
 	filescanner_process_media(file, mtime, size, type, NULL);
 
+	cache_artwork_ping(file, mtime, !is_bulkscan);
+	// TODO [artworkcache] If entry in artwork cache exists for no artwork available, delete the entry if media file has embedded artwork
+
 	counter++;
 
 	/* When in bulk mode, split transaction in pieces of 200 */
@@ -731,6 +885,19 @@ process_file(char *file, time_t mtime, off_t size, int type, int flags)
 	  process_playlist(file, mtime);
 	break;
 
+      case FILE_SMARTPL:
+	DPRINTF(E_DBG, L_SCAN, "Smart playlist file: %s\n", file);
+	scan_smartpl(file, mtime);
+	break;
+
+      case FILE_ARTWORK:
+	DPRINTF(E_DBG, L_SCAN, "Artwork file: %s\n", file);
+	cache_artwork_ping(file, mtime, !is_bulkscan);
+
+	// TODO [artworkcache] If entry in artwork cache exists for no artwork available for a album with files in the same directory, delete the entry
+
+	break;
+
       case FILE_CTRL_REMOTE:
 	remote_pairing_read_pin(file);
 	break;
@@ -753,11 +920,7 @@ process_file(char *file, time_t mtime, off_t size, int type, int flags)
 
 	DPRINTF(E_LOG, L_SCAN, "Startup rescan triggered, found init-rescan file: %s\n", file);
 
-	inofd_event_unset(); // Clears all inotify watches
-	db_watch_clear();
-
-	inofd_event_set();
-	bulk_scan(F_SCAN_BULK | F_SCAN_RESCAN);
+	filescanner_initscan(NULL);
 	break;
 
       case FILE_CTRL_FULLSCAN:
@@ -766,13 +929,7 @@ process_file(char *file, time_t mtime, off_t size, int type, int flags)
 
 	DPRINTF(E_LOG, L_SCAN, "Full rescan triggered, found full-rescan file: %s\n", file);
 
-	player_playback_stop();
-	player_queue_clear();
-	inofd_event_unset(); // Clears all inotify watches
-	db_purge_all(); // Clears files, playlists, playlistitems, inotify and groups
-
-	inofd_event_set();
-	bulk_scan(F_SCAN_BULK);
+	filescanner_fullrescan(NULL);
 	break;
 
       default:
@@ -1002,6 +1159,9 @@ bulk_scan(int flags)
   time_t end;
   int i;
 
+  // Set global flag to avoid queued scan requests
+  scanning = 1;
+
   start = time(NULL);
 
   playlists = NULL;
@@ -1066,12 +1226,16 @@ bulk_scan(int flags)
 
       DPRINTF(E_DBG, L_SCAN, "Purging old database content\n");
       db_purge_cruft(start);
+      cache_artwork_purge_cruft(start);
 
       DPRINTF(E_LOG, L_SCAN, "Bulk library scan completed in %.f sec\n", difftime(end, start));
 
       DPRINTF(E_DBG, L_SCAN, "Running post library scan jobs\n");
       db_hook_post_scan();
     }
+
+  // Set scan in progress flag to FALSE
+  scanning = 0;
 }
 
 
@@ -1322,19 +1486,24 @@ static void
 process_inotify_file(struct watch_info *wi, char *path, struct inotify_event *ie)
 {
   struct stat sb;
+  uint32_t path_hash;
   char *deref = NULL;
   char *file = path;
   int type;
+  int i;
   int ret;
 
   DPRINTF(E_SPAM, L_SCAN, "File event: 0x%x, cookie 0x%x, wd %d\n", ie->mask, ie->cookie, wi->wd);
 
+  path_hash = djb_hash(path, strlen(path));
+
   if (ie->mask & IN_DELETE)
     {
       DPRINTF(E_DBG, L_SCAN, "File deleted: %s\n", path);
 
       db_file_delete_bypath(path);
       db_pl_delete_bypath(path);
+      cache_artwork_delete_by_path(path);
     }
 
   if (ie->mask & IN_MOVED_FROM)
@@ -1347,7 +1516,14 @@ process_inotify_file(struct watch_info *wi, char *path, struct inotify_event *ie
 
   if (ie->mask & IN_ATTRIB)
     {
-      DPRINTF(E_DBG, L_SCAN, "File permissions changed: %s\n", path);
+      DPRINTF(E_DBG, L_SCAN, "File attributes changed: %s\n", path);
+
+      // Ignore the IN_ATTRIB if we just got an IN_CREATE
+      for (i = 0; i < INCOMINGFILES_BUFFER_SIZE; i++)
+	{
+	  if (incomingfiles_buffer[i] == path_hash)
+	    return;
+	}
 
 #ifdef HAVE_EUIDACCESS
       if (euidaccess(path, R_OK) < 0)
@@ -1357,7 +1533,8 @@ process_inotify_file(struct watch_info *wi, char *path, struct inotify_event *ie
 	{
 	  DPRINTF(E_LOG, L_SCAN, "File access to '%s' failed: %s\n", path, strerror(errno));
 
-	  db_file_delete_bypath(path);;
+	  db_file_delete_bypath(path);
+	  cache_artwork_delete_by_path(path);
 	}
       else if ((file_type_get(path) == FILE_REGULAR) && (db_file_id_bypath(path) <= 0)) // TODO Playlists
 	{
@@ -1397,7 +1574,15 @@ process_inotify_file(struct watch_info *wi, char *path, struct inotify_event *ie
 	  return;
 	}
 
-      if (S_ISFIFO(sb.st_mode))
+      // Add to the list of files where we ignore IN_ATTRIB until the file is closed again
+      if (S_ISREG(sb.st_mode))
+	{
+	  DPRINTF(E_SPAM, L_SCAN, "Incoming file created '%s' (%d), index %d\n", path, (int)path_hash, incomingfiles_idx);
+
+	  incomingfiles_buffer[incomingfiles_idx] = path_hash;
+	  incomingfiles_idx = (incomingfiles_idx + 1) % INCOMINGFILES_BUFFER_SIZE;
+	}
+      else if (S_ISFIFO(sb.st_mode))
 	ie->mask |= IN_CLOSE_WRITE;
     }
 
@@ -1405,6 +1590,15 @@ process_inotify_file(struct watch_info *wi, char *path, struct inotify_event *ie
     {
       DPRINTF(E_DBG, L_SCAN, "File closed: %s\n", path);
 
+      // File has been closed so remove from the IN_ATTRIB ignore list
+      for (i = 0; i < INCOMINGFILES_BUFFER_SIZE; i++)
+	if (incomingfiles_buffer[i] == path_hash)
+	  {
+	    DPRINTF(E_SPAM, L_SCAN, "Incoming file closed '%s' (%d), index %d\n", path, (int)path_hash, i);
+
+	    incomingfiles_buffer[i] = 0;
+	  }
+
       ret = lstat(path, &sb);
       if (ret < 0)
 	{
@@ -1806,6 +2000,134 @@ exit_cb(int fd, short event, void *arg)
   scan_exit = 1;
 }
 
+static void
+command_cb(int fd, short what, void *arg)
+{
+  struct filescanner_command *cmd;
+  int ret;
+
+  ret = read(cmd_pipe[0], &cmd, sizeof(cmd));
+  if (ret != sizeof(cmd))
+    {
+      DPRINTF(E_LOG, L_SCAN, "Could not read command! (read %d): %s\n", ret, (ret < 0) ? strerror(errno) : "-no error-");
+      goto readd;
+    }
+
+  if (cmd->nonblock)
+    {
+      cmd->func(cmd);
+
+      free(cmd);
+      goto readd;
+    }
+
+  pthread_mutex_lock(&cmd->lck);
+
+  ret = cmd->func(cmd);
+  cmd->ret = ret;
+
+  pthread_cond_signal(&cmd->cond);
+  pthread_mutex_unlock(&cmd->lck);
+
+ readd:
+  event_add(&cmdev, NULL);
+}
+
+static int
+filescanner_initscan(struct filescanner_command *cmd)
+{
+  DPRINTF(E_LOG, L_SCAN, "Startup rescan triggered\n");
+
+  inofd_event_unset(); // Clears all inotify watches
+  db_watch_clear();
+
+  inofd_event_set();
+  bulk_scan(F_SCAN_BULK | F_SCAN_RESCAN);
+
+  return 0;
+}
+
+static int
+filescanner_fullrescan(struct filescanner_command *cmd)
+{
+  DPRINTF(E_LOG, L_SCAN, "Full rescan triggered\n");
+
+  player_playback_stop();
+  player_queue_clear();
+  inofd_event_unset(); // Clears all inotify watches
+  db_purge_all(); // Clears files, playlists, playlistitems, inotify and groups
+
+  inofd_event_set();
+  bulk_scan(F_SCAN_BULK);
+
+  return 0;
+}
+
+void
+filescanner_trigger_initscan(void)
+{
+  struct filescanner_command *cmd;
+
+  if (scanning)
+    {
+      DPRINTF(E_INFO, L_SCAN, "Scan already running, ignoring request to trigger a new init scan\n");
+      return;
+    }
+
+
+  cmd = (struct filescanner_command *)malloc(sizeof(struct filescanner_command));
+  if (!cmd)
+    {
+      DPRINTF(E_LOG, L_SCAN, "Could not allocate cache_command\n");
+      return;
+    }
+
+  memset(cmd, 0, sizeof(struct filescanner_command));
+
+  cmd->nonblock = 1;
+
+  cmd->func = filescanner_initscan;
+
+  nonblock_command(cmd);
+}
+
+void
+filescanner_trigger_fullrescan(void)
+{
+  struct filescanner_command *cmd;
+
+  if (scanning)
+    {
+      DPRINTF(E_INFO, L_SCAN, "Scan already running, ignoring request to trigger a new init scan\n");
+      return;
+    }
+
+  cmd = (struct filescanner_command *)malloc(sizeof(struct filescanner_command));
+  if (!cmd)
+    {
+      DPRINTF(E_LOG, L_SCAN, "Could not allocate cache_command\n");
+      return;
+    }
+
+  memset(cmd, 0, sizeof(struct filescanner_command));
+
+  cmd->nonblock = 1;
+
+  cmd->func = filescanner_fullrescan;
+
+  nonblock_command(cmd);
+}
+
+/*
+ * Query the status of the filescanner
+ * @return 1 if scan is running, otherwise 0
+ */
+int
+filescanner_scanning(void)
+{
+  return scanning;
+}
+
 /* Thread: main */
 int
 filescanner_init(void)
@@ -1813,6 +2135,7 @@ filescanner_init(void)
   int ret;
 
   scan_exit = 0;
+  scanning = 0;
 
   evbase_scan = event_base_new();
   if (!evbase_scan)
@@ -1858,6 +2181,21 @@ filescanner_init(void)
   event_base_set(evbase_scan, &exitev);
   event_add(&exitev, NULL);
 
+# if defined(__linux__)
+  ret = pipe2(cmd_pipe, O_CLOEXEC);
+# else
+  ret = pipe(cmd_pipe);
+# endif
+  if (ret < 0)
+    {
+      DPRINTF(E_LOG, L_SCAN, "Could not create command pipe: %s\n", strerror(errno));
+      goto cmd_fail;
+    }
+
+  event_set(&cmdev, cmd_pipe[0], EV_READ, command_cb, NULL);
+  event_base_set(evbase_scan, &cmdev);
+  event_add(&cmdev, NULL);
+
   ret = pthread_create(&tid_scan, NULL, filescanner, NULL);
   if (ret != 0)
     {
@@ -1869,6 +2207,9 @@ filescanner_init(void)
   return 0;
 
  thread_fail:
+ cmd_fail:
+  close(cmd_pipe[0]);
+  close(cmd_pipe[1]);
   close(inofd);
  ino_fail:
 #ifdef USE_EVENTFD
@@ -1927,5 +2268,7 @@ filescanner_deinit(void)
   close(exit_pipe[0]);
   close(exit_pipe[1]);
 #endif
+  close(cmd_pipe[0]);
+  close(cmd_pipe[1]);
   event_base_free(evbase_scan);
 }
