@@ -40,8 +40,6 @@
 #include <uninorm.h>
 #include <unistd.h>
 
-#include <avl.h>
-
 #include "logger.h"
 #include "db.h"
 #include "conffile.h"
@@ -52,17 +50,18 @@
 #include "httpd_daap.h"
 #include "daap_query.h"
 #include "dmap_common.h"
-#include "daap_cache.h"
+#include "cache.h"
 
 #ifdef HAVE_LIBEVENT2
+# include <event2/event.h>
 # include <event2/http_struct.h>
+#else
+# include <event.h>
 #endif
 
 /* httpd event base, from httpd.c */
 extern struct event_base *evbase_httpd;
 
-/* First session id to use */
-#define DAAP_BASE_SESSION_ID 100
 /* Session timeout in seconds */
 #define DAAP_SESSION_TIMEOUT 0 /* By default the session never times out */
 /* We announce this timeout to the client when returning server capabilities */
@@ -70,6 +69,8 @@ extern struct event_base *evbase_httpd;
 /* Update requests refresh interval in seconds */
 #define DAAP_UPDATE_REFRESH  0
 
+/* Database number for the Radio item */
+#define DAAP_DB_RADIO 2
 
 struct uri_map {
   regex_t preg;
@@ -80,14 +81,16 @@ struct uri_map {
 struct daap_session {
   int id;
   char *user_agent;
-  struct event timeout;
+  struct event *timeout;
+
+  struct daap_session *next;
 };
 
 struct daap_update_request {
   struct evhttp_request *req;
 
   /* Refresh tiemout */
-  struct event timeout;
+  struct event *timeout;
 
   struct daap_update_request *next;
 };
@@ -107,39 +110,26 @@ static char *default_meta_pl = "dmap.itemid,dmap.itemname,dmap.persistentid,com.
 static char *default_meta_group = "dmap.itemname,dmap.persistentid,daap.songalbumartist";
 
 /* DAAP session tracking */
-static avl_tree_t *daap_sessions;
-static int next_session_id;
+static struct daap_session *daap_sessions;
+static struct timeval daap_session_timeout_tv = { DAAP_SESSION_TIMEOUT, 0 };
 
 /* Update requests */
 static int current_rev;
 static struct daap_update_request *update_requests;
+static struct timeval daap_update_refresh_tv = { DAAP_UPDATE_REFRESH, 0 };
 
 
 /* Session handling */
-static int
-daap_session_compare(const void *aa, const void *bb)
-{
-  struct daap_session *a = (struct daap_session *)aa;
-  struct daap_session *b = (struct daap_session *)bb;
-
-  if (a->id < b->id)
-    return -1;
-
-  if (a->id > b->id)
-    return 1;
-
-  return 0;
-}
-
 static void
-daap_session_free(void *item)
+daap_session_free(struct daap_session *s)
 {
-  struct daap_session *s;
-
-  s = (struct daap_session *)item;
-
-  if (event_initialized(&s->timeout))
-    evtimer_del(&s->timeout);
+#ifdef HAVE_LIBEVENT2
+  if (s->timeout)
+    event_free(s->timeout);
+#else
+  if (s->timeout)
+    free(s->timeout);
+#endif
 
   if (s->user_agent)
     free(s->user_agent);
@@ -148,9 +138,32 @@ daap_session_free(void *item)
 }
 
 static void
-daap_session_kill(struct daap_session *s)
+daap_session_remove(struct daap_session *s)
 {
-  avl_delete(daap_sessions, s);
+  struct daap_session *ptr;
+  struct daap_session *prev;
+
+  prev = NULL;
+  for (ptr = daap_sessions; ptr; ptr = ptr->next)
+    {
+      if (ptr == s)
+	break;
+
+      prev = ptr;
+    }
+
+  if (!ptr)
+    {
+      DPRINTF(E_LOG, L_DAAP, "Error: Request to remove non-existent session. BUG!\n");
+      return;
+    }
+
+  if (!prev)
+    daap_sessions = s->next;
+  else
+    prev->next = s->next;
+
+  daap_session_free(s);
 }
 
 static void
@@ -162,16 +175,27 @@ daap_session_timeout_cb(int fd, short what, void *arg)
 
   DPRINTF(E_DBG, L_DAAP, "Session %d timed out\n", s->id);
 
-  daap_session_kill(s);
+  daap_session_remove(s);
 }
 
 static struct daap_session *
-daap_session_register(const char *user_agent, int request_session_id)
+daap_session_get(int id)
+{
+  struct daap_session *s;
+
+  for (s = daap_sessions; s; s = s->next)
+    {
+      if (id == s->id)
+	return s;
+    }
+
+  return NULL;
+}
+
+static struct daap_session *
+daap_session_add(const char *user_agent, int request_session_id)
 {
-  struct timeval tv;
   struct daap_session *s;
-  avl_node_t *node;
-  int ret;
 
   s = (struct daap_session *)malloc(sizeof(struct daap_session));
   if (!s)
@@ -182,54 +206,39 @@ daap_session_register(const char *user_agent, int request_session_id)
 
   memset(s, 0, sizeof(struct daap_session));
 
-  if (!request_session_id)
+  if (request_session_id)
     {
-      s->id = next_session_id;
-      next_session_id++;
-    }
-  else if (request_session_id < DAAP_BASE_SESSION_ID)
-    {
-      s->id = request_session_id;
-      if (avl_search(daap_sessions, s))
+      if (daap_session_get(request_session_id))
 	{
 	  DPRINTF(E_LOG, L_DAAP, "Session id requested in login (%d) is not available\n", request_session_id);
 	  return NULL;
 	}
-    }    
+      
+      s->id = request_session_id;
+    }
   else
     {
-      DPRINTF(E_LOG, L_DAAP, "Session id requested in login (%d) must be below %d\n", request_session_id, DAAP_BASE_SESSION_ID);
-      return NULL;
+      while ( (s->id = rand() + 100) && daap_session_get(s->id) );
     }
 
   if (user_agent)
     s->user_agent = strdup(user_agent);
 
-  if (DAAP_SESSION_TIMEOUT > 0) {
-    evtimer_set(&s->timeout, daap_session_timeout_cb, s);
-    event_base_set(evbase_httpd, &s->timeout);
-  }
+  if (daap_sessions)
+    s->next = daap_sessions;
 
-  node = avl_insert(daap_sessions, s);
-  if (!node)
+  daap_sessions = s;
+
+#ifdef HAVE_LIBEVENT2
+  if (DAAP_SESSION_TIMEOUT > 0)
     {
-      DPRINTF(E_LOG, L_DAAP, "Could not register DAAP session: %s\n", strerror(errno));
-
-      if (user_agent)
-	free(s->user_agent);
-
-      free(s);
-      return NULL;
+      s->timeout = evtimer_new(evbase_httpd, daap_session_timeout_cb, s);
+      if (!s->timeout)
+	DPRINTF(E_LOG, L_DAAP, "Could not add session timeout event for session %d\n", s->id);
+      else
+        evtimer_add(s->timeout, &daap_session_timeout_tv);
     }
-
-  if (DAAP_SESSION_TIMEOUT > 0) {
-    evutil_timerclear(&tv);
-    tv.tv_sec = DAAP_SESSION_TIMEOUT;
-    
-    ret = evtimer_add(&s->timeout, &tv);
-    if (ret < 0)
-      DPRINTF(E_LOG, L_DAAP, "Could not add session timeout event for session %d\n", s->id);
-  }
+#endif
 
   return s;
 }
@@ -237,11 +246,9 @@ daap_session_register(const char *user_agent, int request_session_id)
 struct daap_session *
 daap_session_find(struct evhttp_request *req, struct evkeyvalq *query, struct evbuffer *evbuf)
 {
-  struct daap_session needle;
-  struct timeval tv;
   struct daap_session *s;
-  avl_node_t *node;
   const char *param;
+  int id;
   int ret;
 
   if (!req)
@@ -254,31 +261,19 @@ daap_session_find(struct evhttp_request *req, struct evkeyvalq *query, struct ev
       goto invalid;
     }
 
-  ret = safe_atoi32(param, &needle.id);
+  ret = safe_atoi32(param, &id);
   if (ret < 0)
     goto invalid;
 
-  node = avl_search(daap_sessions, &needle);
-  if (!node)
+  s = daap_session_get(id);
+  if (!s)
     {
-      DPRINTF(E_WARN, L_DAAP, "DAAP session id %d not found\n", needle.id);
+      DPRINTF(E_WARN, L_DAAP, "DAAP session id %d not found\n", id);
       goto invalid;
     }
 
-  s = (struct daap_session *)node->item;
-
-  if (DAAP_SESSION_TIMEOUT > 0) {
-    event_del(&s->timeout);
-  }
-
-  evutil_timerclear(&tv);
-  tv.tv_sec = DAAP_SESSION_TIMEOUT;
-
-  if (DAAP_SESSION_TIMEOUT > 0) {
-    ret = evtimer_add(&s->timeout, &tv);
-    if (ret < 0)
-      DPRINTF(E_LOG, L_DAAP, "Could not add session timeout event for session %d\n", s->id);
-  }
+  if (s->timeout)
+    evtimer_add(s->timeout, &daap_session_timeout_tv);
 
   return s;
 
@@ -289,6 +284,20 @@ daap_session_find(struct evhttp_request *req, struct evkeyvalq *query, struct ev
 
 
 /* Update requests helpers */
+static void
+update_free(struct daap_update_request *ur)
+{
+#ifdef HAVE_LIBEVENT2
+  if (ur->timeout)
+    event_free(ur->timeout);
+#else
+  if (ur->timeout)
+    free(ur->timeout);
+#endif
+
+  free(ur);
+}
+
 static void
 update_remove(struct daap_update_request *ur)
 {
@@ -309,14 +318,8 @@ update_remove(struct daap_update_request *ur)
 
       p->next = ur->next;
     }
-}
 
-static void
-update_free(struct daap_update_request *ur)
-{
-  if (event_initialized(&ur->timeout))
-    evtimer_del(&ur->timeout);
-  free(ur);
+  update_free(ur);
 }
 
 static void
@@ -356,7 +359,6 @@ update_refresh_cb(int fd, short event, void *arg)
   httpd_send_reply(ur->req, HTTP_OK, "OK", evbuf);
 
   update_remove(ur);
-  update_free(ur);
 }
 
 static void
@@ -374,7 +376,6 @@ update_fail_cb(struct evhttp_connection *evcon, void *arg)
     evhttp_connection_set_closecb(evc, NULL, NULL);
 
   update_remove(ur);
-  update_free(ur);
 }
 
 
@@ -436,7 +437,7 @@ daap_sort_build(struct sort_ctx *ctx, char *str)
   len = strlen(str);
   if (len > 0)
     {
-      ret = u8_normalize(UNINORM_NFD, (uint8_t *)str, len, NULL, &len);
+      ret = u8_normalize(UNINORM_NFD, (uint8_t *)str, len + 1, NULL, &len);
       if (!ret)
 	{
 	  DPRINTF(E_LOG, L_DAAP, "Could not normalize string for sort header\n");
@@ -506,14 +507,31 @@ daap_sort_finalize(struct sort_ctx *ctx)
   dmap_add_int(ctx->headerlist, "mshn", ctx->misc_mshn); /* 12 */
 }
 
+/* Remotes are clients that will issue DACP commands. For these clients we will
+ * do the playback, and we will not stream to them. This is a crude function to
+ * identify them, so we can give them appropriate treatment.
+ */
+static int
+is_remote(const char *user_agent)
+{
+  if (!user_agent)
+    return 0;
+  if (strcasestr(user_agent, "remote"))
+    return 1;
+  if (strstr(user_agent, "Retune"))
+    return 1;
+
+  return 0;
+}
+
 /* We try not to return items that the client cannot play (like Spotify and
  * internet streams in iTunes), or which are inappropriate (like internet streams
- * in the album tab in Remote
+ * in the album tab of remotes)
  */
 static void
 user_agent_filter(const char *user_agent, struct query_params *qp)
 {
-  char *filter;
+  const char *filter;
   char *buffer;
   int len;
 
@@ -523,16 +541,10 @@ user_agent_filter(const char *user_agent, struct query_params *qp)
   // Valgrind doesn't like strlen(filter) below, so instead we allocate 128 bytes
   // to hold the string and the leading " AND ". Remember to adjust the 128 if
   // you define strings here that will be too large for the buffer.
-  if (strcasestr(user_agent, "itunes"))
-    filter = strdup("(f.data_kind = 0)"); // Only real files
-  else if (strcasestr(user_agent, "daap"))
-    filter = strdup("(f.data_kind = 0)"); // Only real files
-  else if (strcasestr(user_agent, "remote"))
-    filter = strdup("(f.data_kind <> 1)"); // No internet radio
-  else if (strcasestr(user_agent, "android"))
-    filter = strdup("(f.data_kind <> 1)"); // No internet radio
+  if (is_remote(user_agent))
+    filter = "(f.data_kind <> 1)"; // No internet radio
   else
-    return;
+    filter = "(f.data_kind = 0)"; // Only real files
 
   if (qp->filter)
     {
@@ -547,8 +559,6 @@ user_agent_filter(const char *user_agent, struct query_params *qp)
     qp->filter = strdup(filter);
 
   DPRINTF(E_DBG, L_DAAP, "SQL filter w/client mod: %s\n", qp->filter);
-
-  free(filter);
 }
 
 /* Returns eg /databases/1/containers from /databases/1/containers?meta=dmap.item... */
@@ -676,7 +686,11 @@ get_query_params(struct evkeyvalq *query, int *sort_headers, struct query_params
 
       qp->filter = daap_query_parse_sql(param);
       if (!qp->filter)
-	DPRINTF(E_LOG, L_DAAP, "Ignoring improper DAAP query\n");
+	DPRINTF(E_LOG, L_DAAP, "Ignoring improper DAAP query: %s\n", param);
+
+      /* iTunes seems to default to this when there is a query (which there is for audiobooks, but not for normal playlists) */
+      if (qp->sort == S_NONE)
+	qp->sort = S_ALBUM;
     }
 }
 
@@ -813,7 +827,15 @@ daap_reply_server_info(struct evhttp_request *req, struct evbuffer *evbuf, char
   dmap_add_int(content, "aeSV", apro);       // com.apple.itunes.music-sharing-version (determines if itunes shows share types)
 
   dmap_add_short(content, "ated", 7);        // daap.supportsextradata
-  dmap_add_short(content, "asgr", 3);        // daap.supportsgroups
+
+  /* Sub-optimal user-agent sniffing to solve the problem that iTunes 12.1
+   * does not work if we announce support for groups.
+   */ 
+  ua = evhttp_find_header(headers, "User-Agent");
+  if (ua && (strncmp(ua, "iTunes", strlen("iTunes")) == 0))
+    dmap_add_short(content, "asgr", 0);      // daap.supportsgroups (1=artists, 2=albums, 3=both)
+  else
+    dmap_add_short(content, "asgr", 3);      // daap.supportsgroups (1=artists, 2=albums, 3=both)
 
 //  dmap_add_long(content, "asse", 0x80000); // unknown - used by iTunes
 
@@ -846,7 +868,10 @@ daap_reply_server_info(struct evhttp_request *req, struct evbuffer *evbuf, char
   dmap_add_char(content, "msix", 1);         // dmap.supportsindex
 //  dmap_add_char(content, "msrs", 1);       // dmap.supportsresolve
 
-  dmap_add_int(content, "msdc", 1);          // dmap.databasescount
+  dmap_add_int(content, "msdc", 2);          // dmap.databasescount
+
+//  dmap_add_int(content, "mstc", );          // dmap.utctime
+//  dmap_add_int(content, "msto", );          // dmap.utcoffset
 
   // Create container
   dmap_add_container(evbuf, "msrv", EVBUFFER_LENGTH(content));
@@ -959,7 +984,7 @@ daap_reply_login(struct evhttp_request *req, struct evbuffer *evbuf, char **uri,
   else
     request_session_id = 0;
 
-  s = daap_session_register(ua, request_session_id);
+  s = daap_session_add(ua, request_session_id);
   if (!s)
     {
       dmap_send_error(req, "mlog", "Could not start session");
@@ -984,7 +1009,7 @@ daap_reply_logout(struct evhttp_request *req, struct evbuffer *evbuf, char **uri
   if (!s)
     return -1;
 
-  daap_session_kill(s);
+  daap_session_remove(s);
 
   httpd_send_reply(req, 204, "Logout Successful", evbuf);
 
@@ -994,7 +1019,6 @@ daap_reply_logout(struct evhttp_request *req, struct evbuffer *evbuf, char **uri
 static int
 daap_reply_update(struct evhttp_request *req, struct evbuffer *evbuf, char **uri, struct evkeyvalq *query, const char *ua)
 {
-  struct timeval tv;
   struct daap_session *s;
   struct daap_update_request *ur;
   struct evhttp_connection *evcon;
@@ -1009,7 +1033,7 @@ daap_reply_update(struct evhttp_request *req, struct evbuffer *evbuf, char **uri
   param = evhttp_find_header(query, "revision-number");
   if (!param)
     {
-      DPRINTF(E_LOG, L_DAAP, "Missing revision-number in client update request\n");
+      DPRINTF(E_DBG, L_DAAP, "Missing revision-number in client update request\n");
       /* Some players (Amarok, Banshee) don't supply a revision number.
 	 They get a standard update of everything. */
       param = "1";  /* Default to "1" will insure update */
@@ -1056,24 +1080,25 @@ daap_reply_update(struct evhttp_request *req, struct evbuffer *evbuf, char **uri
     }
   memset(ur, 0, sizeof(struct daap_update_request));
 
-  if (DAAP_UPDATE_REFRESH > 0) {
-    evtimer_set(&ur->timeout, update_refresh_cb, ur);
-    event_base_set(evbase_httpd, &ur->timeout);
-    
-    evutil_timerclear(&tv);
-    tv.tv_sec = DAAP_UPDATE_REFRESH;
-    
-    ret = evtimer_add(&ur->timeout, &tv);
-    if (ret < 0)
-      {
-	DPRINTF(E_LOG, L_DAAP, "Could not add update timeout event\n");
-	
-	dmap_send_error(req, "mupd", "Could not register timer");
-	
-	update_free(ur);
-	return -1;
-      }
-  }
+#ifdef HAVE_LIBEVENT2
+  if (DAAP_UPDATE_REFRESH > 0)
+    {
+      ur->timeout = evtimer_new(evbase_httpd, update_refresh_cb, ur);
+      if (ur->timeout)
+	ret = evtimer_add(ur->timeout, &daap_update_refresh_tv);
+      else
+	ret = -1;
+
+      if (ret < 0)
+	{
+	  DPRINTF(E_LOG, L_DAAP, "Out of memory for update request event\n");
+
+	  dmap_send_error(req, "mupd", "Could not register timer");	
+	  update_free(ur);
+	  return -1;
+	}
+    }
+#endif
 
   /* NOTE: we may need to keep reqd_rev in there too */
   ur->req = req;
@@ -1104,9 +1129,11 @@ static int
 daap_reply_dblist(struct evhttp_request *req, struct evbuffer *evbuf, char **uri, struct evkeyvalq *query, const char *ua)
 {
   struct evbuffer *content;
+  struct evbuffer *item;
   struct daap_session *s;
   cfg_t *lib;
   char *name;
+  char *name_radio;
   int count;
 
   s = daap_session_find(req, query, evbuf);
@@ -1115,6 +1142,7 @@ daap_reply_dblist(struct evhttp_request *req, struct evbuffer *evbuf, char **uri
 
   lib = cfg_getsec(cfg, "library");
   name = cfg_getstr(lib, "name");
+  name_radio = cfg_getstr(lib, "name_radio");
 
   content = evbuffer_new();
   if (!content)
@@ -1125,29 +1153,66 @@ daap_reply_dblist(struct evhttp_request *req, struct evbuffer *evbuf, char **uri
       return -1;
     }
 
-  dmap_add_int(content, "miid", 1);
-  dmap_add_long(content, "mper", 1);
-  dmap_add_int(content, "mdbk", 1);
-  dmap_add_int(content, "aeCs", 1);
-  dmap_add_string(content, "minm", name);
+  // Add db entry for library with dbid = 1
+  item = evbuffer_new();
+  if (!item)
+    {
+      DPRINTF(E_LOG, L_DAAP, "Could not create evbuffer for DAAP dblist library item\n");
 
+      dmap_send_error(req, "avdb", "Out of memory");
+      return -1;
+    }
+
+  dmap_add_int(item, "miid", 1);
+  dmap_add_long(item, "mper", 1);
+  dmap_add_int(item, "mdbk", 1);
+  dmap_add_int(item, "aeCs", 1);
+  dmap_add_string(item, "minm", name);
   count = db_files_get_count();
-  dmap_add_int(content, "mimc", count);
-
+  dmap_add_int(item, "mimc", count);
   count = db_pl_get_count(); // TODO Don't count empty smart playlists, because they get excluded in aply
-  dmap_add_int(content, "mctc", count);
-
+  dmap_add_int(item, "mctc", count);
 //  dmap_add_int(content, "aeMk", 0x405);   // com.apple.itunes.extended-media-kind (OR of all in library)
-  dmap_add_int(content, "meds", 3);
+  dmap_add_int(item, "meds", 3);
+
+  // Create container for library db
+  dmap_add_container(content, "mlit", EVBUFFER_LENGTH(item));
+  evbuffer_add_buffer(content, item);
+  evbuffer_free(item);
+
+  // Add second db entry for radio with dbid = DAAP_DB_RADIO
+  item =  evbuffer_new();
+  if (!item)
+    {
+      DPRINTF(E_LOG, L_DAAP, "Could not create evbuffer for DAAP dblist radio item\n");
+
+      dmap_send_error(req, "avdb", "Out of memory");
+      return -1;
+    }
+
+  dmap_add_int(item, "miid", DAAP_DB_RADIO);
+  dmap_add_long(item, "mper", DAAP_DB_RADIO);
+  dmap_add_int(item, "mdbk", 0x64);
+  dmap_add_int(item, "aeCs", 0);
+  dmap_add_string(item, "minm", name_radio);
+  count = db_pl_get_count();       // TODO This counts too much, should only include stream playlists
+  dmap_add_int(item, "mimc", count);
+  dmap_add_int(item, "mctc", 0);
+  dmap_add_int(item, "aeMk", 1);   // com.apple.itunes.extended-media-kind (OR of all in library)
+  dmap_add_int(item, "meds", 3);
+
+  // Create container for radio db
+  dmap_add_container(content, "mlit", EVBUFFER_LENGTH(item));
+  evbuffer_add_buffer(content, item);
+  evbuffer_free(item);
 
   // Create container
-  dmap_add_container(evbuf, "avdb", EVBUFFER_LENGTH(content) + 61);
+  dmap_add_container(evbuf, "avdb", EVBUFFER_LENGTH(content) + 53);
   dmap_add_int(evbuf, "mstt", 200);     /* 12 */
   dmap_add_char(evbuf, "muty", 0);      /* 9 */
-  dmap_add_int(evbuf, "mtco", 1);       /* 12 */
-  dmap_add_int(evbuf, "mrco", 1);       /* 12 */
-  dmap_add_container(evbuf, "mlcl", EVBUFFER_LENGTH(content) + 8); /* 8 */
-  dmap_add_container(evbuf, "mlit", EVBUFFER_LENGTH(content));     /* 8 */
+  dmap_add_int(evbuf, "mtco", 2);       /* 12 */
+  dmap_add_int(evbuf, "mrco", 2);       /* 12 */
+  dmap_add_container(evbuf, "mlcl", EVBUFFER_LENGTH(content)); /* 8 */
   evbuffer_add_buffer(evbuf, content);
   evbuffer_free(content);
 
@@ -1169,10 +1234,12 @@ daap_reply_songlist_generic(struct evhttp_request *req, struct evbuffer *evbuf,
   struct sort_ctx *sctx;
   const char *param;
   const char *client_codecs;
+  char *last_codectype;
   char *tag;
   int nmeta;
   int sort_headers;
   int nsongs;
+  int remote;
   int transcode;
   int ret;
 
@@ -1299,19 +1366,40 @@ daap_reply_songlist_generic(struct evhttp_request *req, struct evbuffer *evbuf,
       goto out_query_free;
     }
 
+  remote = is_remote(ua);
+
+  client_codecs = NULL;
+  if (!remote && req)
+    {
+      headers = evhttp_request_get_input_headers(req);
+      client_codecs = evhttp_find_header(headers, "Accept-Codecs");
+    }
+
   nsongs = 0;
+  last_codectype = NULL;
   while (((ret = db_query_fetch_file(&qp, &dbmfi)) == 0) && (dbmfi.id))
     {
       nsongs++;
 
-      client_codecs = NULL;
-      if (req)
+      if (!dbmfi.codectype)
 	{
-	  headers = evhttp_request_get_input_headers(req);
-	  client_codecs = evhttp_find_header(headers, "Accept-Codecs");
-	}
+	  DPRINTF(E_LOG, L_DAAP, "Cannot transcode '%s', codec type is unknown\n", dbmfi.fname);
 
-      transcode = transcode_needed(ua, client_codecs, dbmfi.codectype);
+	  transcode = 0;
+	}
+      else if (remote)
+	{
+	  transcode = 1;
+	}
+      else if (!last_codectype || (strcmp(last_codectype, dbmfi.codectype) != 0))
+	{
+	  transcode = transcode_needed(ua, client_codecs, dbmfi.codectype);
+
+	  if (last_codectype)
+	    free(last_codectype);
+
+	  last_codectype = strdup(dbmfi.codectype);
+	}
 
       ret = dmap_encode_file_metadata(songlist, song, &dbmfi, meta, nmeta, sort_headers, transcode);
       if (ret < 0)
@@ -1339,6 +1427,9 @@ daap_reply_songlist_generic(struct evhttp_request *req, struct evbuffer *evbuf,
 
   DPRINTF(E_DBG, L_DAAP, "Done with song list, %d songs\n", nsongs);
 
+  if (last_codectype)
+    free(last_codectype);
+
   if (nmeta > 0)
     free(meta);
 
@@ -1461,16 +1552,21 @@ daap_reply_playlists(struct evhttp_request *req, struct evbuffer *evbuf, char **
   struct daap_session *s;
   struct evbuffer *playlistlist;
   struct evbuffer *playlist;
+  cfg_t *lib;
   const struct dmap_field_map *dfm;
   const struct dmap_field *df;
   const struct dmap_field **meta;
   const char *param;
   char **strval;
+  int database;
+  int cfg_radiopl;
   int nmeta;
   int npls;
   int32_t plid;
   int32_t pltype;
   int32_t plitems;
+  int32_t plstreams;
+  int32_t plparent;
   int i;
   int ret;
 
@@ -1478,6 +1574,17 @@ daap_reply_playlists(struct evhttp_request *req, struct evbuffer *evbuf, char **
   if (!s)
     return -1;
 
+  ret = safe_atoi32(uri[1], &database);
+  if (ret < 0)
+    {
+      dmap_send_error(req, "aply", "Invalid database ID");
+
+      return -1;
+    }
+
+  lib = cfg_getsec(cfg, "library");
+  cfg_radiopl = cfg_getbool(lib, "radio_playlists");
+
   ret = evbuffer_expand(evbuf, 61);
   if (ret < 0)
     {
@@ -1570,8 +1677,20 @@ daap_reply_playlists(struct evhttp_request *req, struct evbuffer *evbuf, char **
       if (safe_atoi32(dbpli.items, &plitems) != 0)
 	continue;
 
-      /* Don't add empty smart playlists */
-      if ((plid > 1) && (pltype == 1) && (plitems == 0))
+      plstreams = 0;
+      if (safe_atoi32(dbpli.streams, &plstreams) != 0)
+	continue;
+
+      /* Database DAAP_DB_RADIO is radio, so for that db skip playlists without
+       * streams and for other databases skip playlists which are just streams
+       */
+      if ((database == DAAP_DB_RADIO) && (plstreams == 0))
+	continue;
+      if (!cfg_radiopl && (database != DAAP_DB_RADIO) && (plstreams > 0) && (plstreams == plitems))
+	continue;
+
+      /* Don't add empty Special playlists */
+      if ((plid > 1) && (plitems == 0) && (pltype == PL_SPECIAL))
 	continue;
 
       npls++;
@@ -1585,14 +1704,20 @@ daap_reply_playlists(struct evhttp_request *req, struct evbuffer *evbuf, char **
 	  if (dfm == &dfm_dmap_mimc)
 	    continue;
 
-	  /* com.apple.itunes.smart-playlist - type = 1 AND id != 1 */
+	  /* Add field "com.apple.itunes.smart-playlist" for special and smart playlists
+	     (excluding the special playlist for "library" with id = 1) */
 	  if (dfm == &dfm_dmap_aeSP)
 	    {
-	      if ((pltype == PL_SMART) && (plid != 1))
+	      if ((pltype == PL_SMART) || ((pltype == PL_SPECIAL) && (plid != 1)))
+		{
+		  dmap_add_char(playlist, "aeSP", 1);
+		}
+
+	      /* Add field "com.apple.itunes.special-playlist" for special playlists
+		 (excluding the special playlist for "library" with id = 1) */
+	      if ((pltype == PL_SPECIAL) && (plid != 1))
 		{
 		  int32_t aePS = 0;
-		  dmap_add_char(playlist, "aeSP", 1);
-
 		  ret = safe_atoi32(dbpli.special_id, &aePS);
 		  if ((ret == 0) && (aePS > 0))
 		    dmap_add_char(playlist, "aePS", aePS);
@@ -1616,11 +1741,14 @@ daap_reply_playlists(struct evhttp_request *req, struct evbuffer *evbuf, char **
 	}
 
       /* Item count (mimc) */
-      if (plitems > 0)
-	dmap_add_int(playlist, "mimc", plitems);
+      dmap_add_int(playlist, "mimc", plitems);
 
       /* Container ID (mpco) */
-      dmap_add_int(playlist, "mpco", 0);
+      ret = safe_atoi32(dbpli.parent_id, &plparent);
+      if (ret == 0)
+	dmap_add_int(playlist, "mpco", plparent);
+      else
+	dmap_add_int(playlist, "mpco", 0);
 
       /* Base playlist (abpl), id = 1 */
       if (plid == 1)
@@ -1894,7 +2022,7 @@ daap_reply_groups(struct evhttp_request *req, struct evbuffer *evbuf, char **uri
       if ((ret == 0) && (val > 0))
 	dmap_add_int(group, "miid", val);
 
-      DPRINTF(E_DBG, L_DAAP, "Done with group\n");
+      DPRINTF(E_SPAM, L_DAAP, "Done with group\n");
 
       dmap_add_container(grouplist, "mlit", EVBUFFER_LENGTH(group));
       ret = evbuffer_add_buffer(grouplist, group);
@@ -1913,7 +2041,10 @@ daap_reply_groups(struct evhttp_request *req, struct evbuffer *evbuf, char **uri
   evbuffer_free(group);
 
   if (qp.filter)
-    free(qp.filter);
+    {
+        free(qp.filter);
+        qp.filter = NULL;
+    }
 
   if (ret < 0)
     {
@@ -1993,8 +2124,7 @@ daap_reply_groups(struct evhttp_request *req, struct evbuffer *evbuf, char **uri
   evbuffer_free(grouplist);
 
  out_qfilter_free:
-  if (qp.filter)
-    free(qp.filter);
+  free(qp.filter);
 
   return -1;
 }
@@ -2039,11 +2169,13 @@ daap_reply_browse(struct evhttp_request *req, struct evbuffer *evbuf, char **uri
     {
       tag = "abgn";
       qp.type = Q_BROWSE_GENRES;
+      qp.sort = S_GENRE;
     }
   else if (strcmp(uri[3], "composers") == 0)
     {
       tag = "abcp";
       qp.type = Q_BROWSE_COMPOSERS;
+      qp.sort = S_COMPOSER;
     }
   else
     {
@@ -2255,9 +2387,9 @@ daap_reply_extra_data(struct evhttp_request *req, struct evbuffer *evbuf, char *
     }
 
   if (strcmp(uri[2], "groups") == 0)
-    ret = artwork_get_group(id, max_w, max_h, ART_CAN_PNG | ART_CAN_JPEG, evbuf);
+    ret = artwork_get_group(evbuf, id, max_w, max_h);
   else if (strcmp(uri[2], "items") == 0)
-    ret = artwork_get_item(id, max_w, max_h, ART_CAN_PNG | ART_CAN_JPEG, evbuf);
+    ret = artwork_get_item(evbuf, id, max_w, max_h);
 
   switch (ret)
     {
@@ -2294,14 +2426,9 @@ daap_reply_extra_data(struct evhttp_request *req, struct evbuffer *evbuf, char *
 static int
 daap_stream(struct evhttp_request *req, struct evbuffer *evbuf, char **uri, struct evkeyvalq *query, const char *ua)
 {
-  struct daap_session *s;
   int id;
   int ret;
 
-  s = daap_session_find(req, query, evbuf);
-  if (!s)
-    return -1;
-
   ret = safe_atoi32(uri[3], &id);
   if (ret < 0)
     evhttp_send_error(req, HTTP_BADREQUEST, "Bad Request");
@@ -2311,6 +2438,23 @@ daap_stream(struct evhttp_request *req, struct evbuffer *evbuf, char **uri, stru
   return ret;
 }
 
+static char *
+uri_relative(char *uri, const char *protocol)
+{
+  char *ret;
+
+  if (strncmp(uri, protocol, strlen(protocol)) != 0)
+    return NULL;
+
+  ret = strchr(uri + strlen(protocol), '/');
+  if (!ret)
+    {
+      DPRINTF(E_LOG, L_DAAP, "Malformed DAAP Request URI '%s'\n", uri);
+      return NULL;
+    }
+
+  return ret;
+}
 
 static char *
 daap_fix_request_uri(struct evhttp_request *req, char *uri)
@@ -2319,25 +2463,21 @@ daap_fix_request_uri(struct evhttp_request *req, char *uri)
 
   /* iTunes 9 gives us an absolute request-uri like
    *  daap://10.1.1.20:3689/server-info
+   * iTunes 12.1 gives us an absolute request-uri for streaming like
+   *  http://10.1.1.20:3689/databases/1/items/1.mp3
    */
 
-  if (strncmp(uri, "daap://", strlen("daap://")) != 0)
-    return uri;
-
-  /* Clear the proxy request flag set by evhttp
-   * due to the request URI being absolute.
-   * It has side-effects on Connection: keep-alive
-   */
-  req->flags &= ~EVHTTP_PROXY_REQUEST;
-
-  ret = strchr(uri + strlen("daap://"), '/');
-  if (!ret)
+  if ( (ret = uri_relative(uri, "daap://")) || (ret = uri_relative(uri, "http://")) )
     {
-      DPRINTF(E_LOG, L_DAAP, "Malformed DAAP Request URI '%s'\n", uri);
-      return NULL;
+      /* Clear the proxy request flag set by evhttp
+       * due to the request URI being absolute.
+       * It has side-effects on Connection: keep-alive
+       */
+      req->flags &= ~EVHTTP_PROXY_REQUEST;
+      return ret;
     }
 
-  return ret;
+  return uri;
 }
 
 
@@ -2668,19 +2808,6 @@ daap_request(struct evhttp_request *req)
    */
   evhttp_add_header(headers, "Content-Type", "application/x-dmap-tagged");
 
-  // Try the cache
-  evbuf = daapcache_get(full_uri);
-  if (evbuf)
-    {
-      httpd_send_reply(req, HTTP_OK, "OK", evbuf); // TODO not all want this reply
-
-      evbuffer_free(evbuf);
-      free(uri);
-      free(full_uri);
-      return;
-    }
-
-  // No cache, so prepare handler arguments and send to the handler
   evbuf = evbuffer_new();
   if (!evbuf)
     {
@@ -2693,6 +2820,19 @@ daap_request(struct evhttp_request *req)
       return;
     }
 
+  // Try the cache
+  ret = cache_daap_get(full_uri, evbuf);
+  if (ret == 0)
+    {
+      httpd_send_reply(req, HTTP_OK, "OK", evbuf); // TODO not all want this reply
+
+      evbuffer_free(evbuf);
+      free(uri);
+      free(full_uri);
+      return;
+    }
+
+  // No cache, so prepare handler arguments and send to the handler
   evhttp_parse_query(full_uri, &query);
 
   clock_gettime(CLOCK_MONOTONIC, &start);
@@ -2703,10 +2843,10 @@ daap_request(struct evhttp_request *req)
 
   msec = (end.tv_sec * 1000 + end.tv_nsec / 1000000) - (start.tv_sec * 1000 + start.tv_nsec / 1000000);
 
-  DPRINTF(E_DBG, L_DB, "DAAP request handled in %d milliseconds\n", msec);
+  DPRINTF(E_DBG, L_DAAP, "DAAP request handled in %d milliseconds\n", msec);
 
-  if (msec > daapcache_threshold())
-    daapcache_add(full_uri, ua, msec);
+  if (msec > cache_daap_threshold())
+    cache_daap_add(full_uri, ua, msec);
 
   evhttp_clear_headers(&query);
   evbuffer_free(evbuf);
@@ -2834,7 +2974,7 @@ daap_init(void)
   int i;
   int ret;
 
-  next_session_id = DAAP_BASE_SESSION_ID;
+  srand((unsigned)time(NULL));
   current_rev = 2;
   update_requests = NULL;
 
@@ -2850,26 +2990,13 @@ daap_init(void)
         }
     }
 
-  daap_sessions = avl_alloc_tree(daap_session_compare, daap_session_free);
-  if (!daap_sessions)
-    {
-      DPRINTF(E_FATAL, L_DAAP, "DAAP init could not allocate DAAP sessions AVL tree\n");
-
-      goto daap_avl_alloc_fail;
-    }
-
   return 0;
-
- daap_avl_alloc_fail:
-  for (i = 0; daap_handlers[i].handler; i++)
-    regfree(&daap_handlers[i].preg);
-
-  return -1;
 }
 
 void
 daap_deinit(void)
 {
+  struct daap_session *s;
   struct daap_update_request *ur;
   struct evhttp_connection *evcon;
   int i;
@@ -2877,7 +3004,11 @@ daap_deinit(void)
   for (i = 0; daap_handlers[i].handler; i++)
     regfree(&daap_handlers[i].preg);
 
-  avl_free_tree(daap_sessions);
+  for (s = daap_sessions; daap_sessions; s = daap_sessions)
+    {
+      daap_sessions = s->next;
+      daap_session_free(s);
+    }
 
   for (ur = update_requests; update_requests; ur = update_requests)
     {
