@@ -46,6 +46,7 @@
 #include "dmap_common.h"
 #include "db.h"
 #include "player.h"
+#include "listener.h"
 
 /* httpd event base, from httpd.c */
 extern struct event_base *evbase_httpd;
@@ -178,7 +179,11 @@ dacp_playingtime(struct evbuffer *evbuf, struct player_status *status, struct me
   if ((status->status == PLAY_STOPPED) || !mfi)
     return;
 
-  dmap_add_int(evbuf, "cant", mfi->song_length - status->pos_ms); /* Remaining time in ms */
+  if (mfi->song_length)
+    dmap_add_int(evbuf, "cant", mfi->song_length - status->pos_ms); /* Remaining time in ms */
+  else
+    dmap_add_int(evbuf, "cant", 0); /* Unknown remaining time */
+
   dmap_add_int(evbuf, "cast", mfi->song_length); /* Song length in ms */
 }
 
@@ -338,10 +343,14 @@ playstatusupdate_cb(int fd, short what, void *arg)
 
 /* Thread: player */
 static void
-dacp_playstatus_update_handler(void)
+dacp_playstatus_update_handler(enum listener_event_type type)
 {
   int ret;
 
+  // Only send status update on player change events
+  if (type != LISTENER_PLAYER)
+    return;
+
 #ifdef USE_EVENTFD
   ret = eventfd_write(update_efd, 1);
   if (ret < 0)
@@ -638,10 +647,13 @@ dacp_propset_userrating(const char *value, struct evkeyvalq *query)
       return;
     }
 
-  param = evhttp_find_header(query, "item-spec");
+  param = evhttp_find_header(query, "item-spec"); // Remote
+  if (!param)
+    param = evhttp_find_header(query, "song-spec"); // Retune
+
   if (!param)
     {
-      DPRINTF(E_LOG, L_DACP, "Missing item-spec parameter in dacp.userrating query\n");
+      DPRINTF(E_LOG, L_DACP, "Missing item-spec/song-spec parameter in dacp.userrating query\n");
 
       return;
     }
@@ -649,7 +661,7 @@ dacp_propset_userrating(const char *value, struct evkeyvalq *query)
   param = strchr(param, ':');
   if (!param)
     {
-      DPRINTF(E_LOG, L_DACP, "Malformed item-spec parameter in dacp.userrating query\n");
+      DPRINTF(E_LOG, L_DACP, "Malformed item-spec/song-spec parameter in dacp.userrating query\n");
 
       return;
     }
@@ -658,7 +670,7 @@ dacp_propset_userrating(const char *value, struct evkeyvalq *query)
   ret = safe_hextou32(param, &itemid);
   if (ret < 0)
     {
-      DPRINTF(E_LOG, L_DACP, "Couldn't convert item-spec to an integer in dacp.userrating (%s)\n", param);
+      DPRINTF(E_LOG, L_DACP, "Couldn't convert item-spec/song-spec to an integer in dacp.userrating (%s)\n", param);
 
       return;
     }
@@ -722,7 +734,10 @@ dacp_reply_cue_play(struct evhttp_request *req, struct evbuffer *evbuf, char **u
   const char *cuequery;
   const char *param;
   uint32_t id;
+  uint32_t pos;
   int clear;
+  struct player_history *history;
+  int hist;
   int ret;
 
   /* /cue?command=play&query=...&sort=...&index=N */
@@ -770,19 +785,52 @@ dacp_reply_cue_play(struct evhttp_request *req, struct evbuffer *evbuf, char **u
     dacp_propset_shufflestate(param, NULL);
 
   id = 0;
+  pos = 0;
   param = evhttp_find_header(query, "index");
   if (param)
     {
-      ret = safe_atou32(param, &id);
+      ret = safe_atou32(param, &pos);
       if (ret < 0)
 	DPRINTF(E_LOG, L_DACP, "Invalid index (%s) in cue request\n", param);
     }
 
-  /* If selection was from Up Next queue (command will be playnow), then index is relative */
+  /* If selection was from Up Next queue or history queue (command will be playnow), then index is relative */
+  hist = 0;
   if ((param = evhttp_find_header(query, "command")) && (strcmp(param, "playnow") == 0))
-    id += status.pos_pl;
+    {
+      /* If mode parameter is -1, the index is relative to the history queue, otherwise to the Up Next queue */
+      param = evhttp_find_header(query, "mode");
+      if (param && (strcmp(param, "-1") == 0))
+	{
+	  /* Play from history queue */
+	  hist = 1;
+	  history = player_history_get();
+	  if (history->count > pos)
+	    {
+	      pos = (history->start_index + history->count - pos - 1) % MAX_HISTORY_COUNT;
+	      id = history->id[pos];
+	    }
+	  else
+	    {
+	      DPRINTF(E_LOG, L_DACP, "Could not start playback from history\n");
+
+	      dmap_send_error(req, "cacr", "Playback failed to start");
+	      return;
+	    }
+	}
+      else
+	{
+	  /* Play from Up Next queue */
+	  pos += status.pos_pl;
+	}
+    }
+
+  /* If playing from history queue, the pos holds the id of the item to play */
+  if (hist)
+    ret = player_playback_startid(id, &id);
+  else
+    ret = player_playback_startpos(pos, &id);
 
-  ret = player_playback_start(&id);
   if (ret < 0)
     {
       DPRINTF(E_LOG, L_DACP, "Could not start playback\n");
@@ -856,6 +904,7 @@ dacp_reply_playspec(struct evhttp_request *req, struct evbuffer *evbuf, char **u
   const char *shuffle;
   uint32_t plid;
   uint32_t id;
+  uint32_t pos;
   int ret;
 
   /* /ctrl-int/1/playspec?database-spec='dmap.persistentid:0x1'&container-spec='dmap.persistentid:0x5'&container-item-spec='dmap.containeritemid:0x9'
@@ -918,7 +967,7 @@ dacp_reply_playspec(struct evhttp_request *req, struct evbuffer *evbuf, char **u
 	}
       param++;
 
-      ret = safe_hextou32(param, &id);
+      ret = safe_hextou32(param, &pos);
       if (ret < 0)
 	{
 	  DPRINTF(E_LOG, L_DACP, "Couldn't convert container-item-spec/item-spec to an integer in playspec (%s)\n", param);
@@ -927,11 +976,11 @@ dacp_reply_playspec(struct evhttp_request *req, struct evbuffer *evbuf, char **u
 	}
     }
   else
-    id = 0;
+    pos = 0;
 
-  DPRINTF(E_DBG, L_DACP, "Playspec request for playlist %d, start song id %d%s\n", plid, id, (shuffle) ? ", shuffle" : "");
+  DPRINTF(E_DBG, L_DACP, "Playspec request for playlist %d, start song id %d%s\n", plid, pos, (shuffle) ? ", shuffle" : "");
 
-  ps = player_queue_make_pl(plid, &id);
+  ps = player_queue_make_pl(plid, &pos);
   if (!ps)
     {
       DPRINTF(E_LOG, L_DACP, "Could not build song queue from playlist %d\n", plid);
@@ -939,7 +988,7 @@ dacp_reply_playspec(struct evhttp_request *req, struct evbuffer *evbuf, char **u
       goto out_fail;
     }
 
-  DPRINTF(E_DBG, L_DACP, "Playspec start song index is %d\n", id);
+  DPRINTF(E_DBG, L_DACP, "Playspec start song index is %d\n", pos);
 
   player_get_status(&status);
 
@@ -953,7 +1002,7 @@ dacp_reply_playspec(struct evhttp_request *req, struct evbuffer *evbuf, char **u
   if (shuffle)
     dacp_propset_shufflestate(shuffle, NULL);
 
-  ret = player_playback_start(&id);
+  ret = player_playback_startpos(pos, &id);
   if (ret < 0)
     {
       DPRINTF(E_LOG, L_DACP, "Could not start playback\n");
@@ -1126,15 +1175,6 @@ dacp_reply_playresume(struct evhttp_request *req, struct evbuffer *evbuf, char *
   evhttp_send_reply(req, HTTP_NOCONTENT, "No Content", evbuf);
 }
 
-static struct player_source *
-next_ps(struct player_source *ps, char shuffle)
-{
-  if (shuffle)
-    return ps->shuffle_next;
-  else
-    return ps->pl_next;
-}
-
 static int
 playqueuecontents_add_source(struct evbuffer *songlist, uint32_t source_id, int pos_in_queue, uint32_t plid)
 {
@@ -1194,10 +1234,9 @@ dacp_reply_playqueuecontents(struct evhttp_request *req, struct evbuffer *evbuf,
   struct daap_session *s;
   struct evbuffer *songlist;
   struct evbuffer *playlists;
-  struct player_source *ps;
-  struct player_source *head;
   struct player_status status;
   struct player_history *history;
+  struct player_queue *queue;
   const char *param;
   int span;
   int i;
@@ -1268,19 +1307,13 @@ dacp_reply_playqueuecontents(struct evhttp_request *req, struct evbuffer *evbuf,
       /* Get queue and make songlist only if playing or paused */
       if (status.status != PLAY_STOPPED)
 	{
-	  /* Fast forward to song currently being played */
-	  head = player_queue_get();
-	  if (head)
+	  queue = player_queue_get(-1, abs(span), status.shuffle);
+	  if (queue)
 	    {
-	      ps = head;
-	      while ((ps->id != status.id) && (ps = next_ps(ps, status.shuffle)) && (ps != head))
-		i++;
-
-	      while ((n < abs(span)) && (ps = next_ps(ps, status.shuffle)) && (ps != head))
+	      i = queue->start_pos;
+	      for (n = 0; (n < queue->count) && (n < abs(span)); n++)
 		{
-		  n++;
-
-		  ret = playqueuecontents_add_source(songlist, ps->id, (n + i + 1), status.plid);
+		  ret = playqueuecontents_add_source(songlist, queue->queue[n], (n + i + 1), status.plid);
 		  if (ret < 0)
 		    {
 		      DPRINTF(E_LOG, L_DACP, "Could not add song to songlist for playqueue-contents\n");
@@ -1290,6 +1323,7 @@ dacp_reply_playqueuecontents(struct evhttp_request *req, struct evbuffer *evbuf,
 		    }
 		}
 	    }
+	  queue_free(queue);
 	}
     }
 
@@ -1403,6 +1437,8 @@ dacp_reply_playqueueedit_add(struct evhttp_request *req, struct evbuffer *evbuf,
   // -> mode=0: add to end of playqueue
   //?command=add&query='dmap.itemid:306'&queuefilter=album:6525753023700533274&sort=album&mode=1&session-id=100
   // -> mode 1: stop playblack, clear playqueue, add songs to playqueue
+  //?command=add&query='dmap.itemid:2'&query-modifier=containers&sort=name&mode=2&session-id=100
+  // -> mode 2: stop playblack, clear playqueue, add shuffled songs from playlist=itemid to playqueue
 
   struct player_source *ps;
   const char *editquery;
@@ -1508,7 +1544,7 @@ dacp_reply_playqueueedit_add(struct evhttp_request *req, struct evbuffer *evbuf,
     }
 
   DPRINTF(E_DBG, L_DACP, "Song queue built, playback starting at index %" PRIu32 "\n", idx);
-  ret = player_playback_start(&idx);
+  ret = player_playback_startpos(idx, NULL);
   if (ret < 0)
     {
       DPRINTF(E_LOG, L_DACP, "Could not start playback\n");
@@ -1802,7 +1838,7 @@ dacp_reply_nowplayingartwork(struct evhttp_request *req, struct evbuffer *evbuf,
   if (ret < 0)
     goto no_artwork;
 
-  ret = artwork_get_item(id, max_w, max_h, ART_CAN_PNG | ART_CAN_JPEG, evbuf);
+  ret = artwork_get_item(evbuf, id, max_w, max_h);
   switch (ret)
     {
       case ART_FMT_PNG:
@@ -2091,12 +2127,8 @@ dacp_reply_setspeakers(struct evhttp_request *req, struct evbuffer *evbuf, char
   do
     {
       param++;
-      /* Hyperfine Remote for Android will send in decimal, others use hex */
-      if ((strlen(param) > 1) && (param[1] != 'x'))
-	ret = safe_atou64(param, &ids[i]);
-      else
-	ret = safe_hextou64(param, &ids[i]);
 
+      ret = safe_hextou64(param, &ids[i]);
       if (ret < 0)
 	{
 	  DPRINTF(E_LOG, L_DACP, "Invalid speaker id in request: %s\n", param);
@@ -2396,7 +2428,7 @@ dacp_init(void)
   event_base_set(evbase_httpd, &updateev);
   event_add(&updateev, NULL);
 
-  player_set_update_handler(dacp_playstatus_update_handler);
+  listener_add(dacp_playstatus_update_handler, LISTENER_PLAYER);
 
   return 0;
 
@@ -2417,7 +2449,7 @@ dacp_deinit(void)
   struct evhttp_connection *evcon;
   int i;
 
-  player_set_update_handler(NULL);
+  listener_remove(dacp_playstatus_update_handler);
 
   for (i = 0; dacp_handlers[i].handler; i++)
     regfree(&dacp_handlers[i].preg);
