@@ -9,28 +9,30 @@
 #define L_DAAP    1
 #define L_DB      2
 #define L_HTTPD   3
-#define L_MAIN    4
-#define L_MDNS    5
-#define L_MISC    6
-#define L_RSP     7
-#define L_SCAN    8
-#define L_XCODE   9
+#define L_HTTP    4
+#define L_MAIN    5
+#define L_MDNS    6
+#define L_MISC    7
+#define L_RSP     8
+#define L_SCAN    9
+#define L_XCODE   10
 /* libevent logging */
-#define L_EVENT   10
-#define L_REMOTE  11
-#define L_DACP    12
-#define L_FFMPEG  13
-#define L_ART     14
-#define L_PLAYER  15
-#define L_RAOP    16
-#define L_LAUDIO  17
-#define L_DMAP    18
-#define L_DBPERF  19
-#define L_SPOTIFY 20
-#define L_LASTFM  21
-#define L_DCACHE  22
+#define L_EVENT   11
+#define L_REMOTE  12
+#define L_DACP    13
+#define L_FFMPEG  14
+#define L_ART     15
+#define L_PLAYER  16
+#define L_RAOP    17
+#define L_LAUDIO  18
+#define L_DMAP    19
+#define L_DBPERF  20
+#define L_SPOTIFY 21
+#define L_LASTFM  22
+#define L_CACHE   23
+#define L_MPD     24
 
-#define N_LOGDOMAINS  23
+#define N_LOGDOMAINS  25
 
 /* Severities */
 #define E_FATAL   0
