@@ -58,15 +58,14 @@ GCRY_THREAD_OPTION_PTHREAD_IMPL;
 #include "db.h"
 #include "logger.h"
 #include "misc.h"
-#include "daap_cache.h"
+#include "cache.h"
 #include "filescanner.h"
 #include "httpd.h"
+#include "mpd.h"
 #include "mdns.h"
 #include "remote_pairing.h"
 #include "player.h"
-#if LIBAVFORMAT_VERSION_MAJOR < 53
-# include "ffmpeg_url_evbuffer.h"
-#endif
+#include "worker.h"
 
 #ifdef LASTFM
 # include "lastfm.h"
@@ -268,11 +267,11 @@ register_services(char *ffid, int no_rsp, int no_daap)
 
   port = cfg_getint(lib, "port");
 
-  /* Register web server service */
-  ret = mdns_register(libname, "_http._tcp", port, txtrecord);
+  /* Register web server service - disabled since we have no web interface */
+/*  ret = mdns_register(libname, "_http._tcp", port, txtrecord);
   if (ret < 0)
     return ret;
-
+*/
   /* Register RSP service */
   if (!no_rsp)
     {
@@ -590,9 +589,6 @@ main(int argc, char **argv)
   avformat_network_init();
 #endif
   av_log_set_callback(logger_ffmpeg);
-#if LIBAVFORMAT_VERSION_MAJOR < 53
-  register_ffmpeg_evbuffer_url_protocol();
-#endif
 
   /* Initialize libgcrypt */
   gcry_control(GCRYCTL_SET_THREAD_CBS, &gcry_threads_pthread);
@@ -675,11 +671,21 @@ main(int argc, char **argv)
       goto db_fail;
     }
 
-  /* Spawn DAAP cache thread */
-  ret = daapcache_init();
+  /* Spawn worker thread */
+  ret = worker_init();
   if (ret != 0)
     {
-      DPRINTF(E_FATAL, L_MAIN, "DAAP cache thread failed to start\n");
+      DPRINTF(E_FATAL, L_MAIN, "Worker thread failed to start\n");
+
+      ret = EXIT_FAILURE;
+      goto worker_fail;
+    }
+
+  /* Spawn cache thread */
+  ret = cache_init();
+  if (ret != 0)
+    {
+      DPRINTF(E_FATAL, L_MAIN, "Cache thread failed to start\n");
 
       ret = EXIT_FAILURE;
       goto cache_fail;
@@ -724,6 +730,18 @@ main(int argc, char **argv)
       goto httpd_fail;
     }
 
+#ifdef MPD
+  /* Spawn MPD thread */
+  ret = mpd_init();
+  if (ret != 0)
+    {
+      DPRINTF(E_FATAL, L_MAIN, "MPD thread failed to start\n");
+
+      ret = EXIT_FAILURE;
+      goto mpd_fail;
+    }
+#endif
+
   /* Start Remote pairing service */
   ret = remote_pairing_init();
   if (ret != 0)
@@ -808,6 +826,13 @@ main(int argc, char **argv)
   httpd_deinit();
 
  httpd_fail:
+  DPRINTF(E_LOG, L_MAIN, "TCPd deinit\n");
+#ifdef MPD
+  DPRINTF(E_LOG, L_MAIN, "MPD deinit\n");
+  mpd_deinit();
+ mpd_fail:
+#endif
+
   DPRINTF(E_LOG, L_MAIN, "Player deinit\n");
   player_deinit();
 
@@ -824,10 +849,14 @@ main(int argc, char **argv)
   filescanner_deinit();
 
  filescanner_fail:
-  DPRINTF(E_LOG, L_MAIN, "DAAP cache deinit\n");
-  daapcache_deinit();
+  DPRINTF(E_LOG, L_MAIN, "Cache deinit\n");
+  cache_deinit();
 
  cache_fail:
+  DPRINTF(E_LOG, L_MAIN, "Worker deinit\n");
+  worker_deinit();
+
+ worker_fail:
   DPRINTF(E_LOG, L_MAIN, "Database deinit\n");
   db_perthread_deinit();
   db_deinit();
