@@ -751,8 +751,9 @@ raop_metadata_prune(uint64_t rtptime)
     }
 }
 
-static struct raop_metadata *
-raop_metadata_prepare(int id, uint64_t rtptime)
+/* Thread: worker */
+struct raop_metadata *
+raop_metadata_prepare(int id)
 {
   struct query_params qp;
   struct db_media_file_info dbmfi;
@@ -772,6 +773,28 @@ raop_metadata_prepare(int id, uint64_t rtptime)
 
   memset(rmd, 0, sizeof(struct raop_metadata));
 
+  /* Get artwork */
+  rmd->artwork = evbuffer_new();
+  if (!rmd->artwork)
+    {
+      DPRINTF(E_LOG, L_RAOP, "Out of memory for artwork evbuffer; no artwork will be sent\n");
+
+      goto skip_artwork;
+    }
+
+  ret = artwork_get_item(rmd->artwork, id, 600, 600);
+  if (ret < 0)
+    {
+      DPRINTF(E_INFO, L_RAOP, "Failed to retrieve artwork for file id %d; no artwork will be sent\n", id);
+
+      evbuffer_free(rmd->artwork);
+      rmd->artwork = NULL;
+    }
+
+  rmd->artwork_fmt = ret;
+
+ skip_artwork:
+
   /* Get dbmfi */
   memset(&qp, 0, sizeof(struct query_params));
   qp.type = Q_ITEMS;
@@ -839,40 +862,11 @@ raop_metadata_prepare(int id, uint64_t rtptime)
       goto out_metadata;
     }
 
-  rmd->start = rtptime;
-  rmd->end = rtptime + (duration * 44100UL) / 1000UL;
-
-  /* Get artwork */
-  rmd->artwork = evbuffer_new();
-  if (!rmd->artwork)
-    {
-      DPRINTF(E_LOG, L_RAOP, "Out of memory for artwork evbuffer; no artwork will be sent\n");
-
-      goto skip_artwork;
-    }
-
-  ret = artwork_get_item(id, 600, 600, ART_CAN_PNG | ART_CAN_JPEG, rmd->artwork);
-  if (ret < 0)
-    {
-      DPRINTF(E_INFO, L_RAOP, "Failed to retrieve artwork for '%s' (%d); no artwork will be sent\n", dbmfi.title, id);
-
-      evbuffer_free(rmd->artwork);
-      rmd->artwork = NULL;
-    }
-
-  rmd->artwork_fmt = ret;
-
- skip_artwork:
   db_query_end(&qp);
 
-  /* Add rmd to metadata list */
-  if (metadata_tail)
-    metadata_tail->next = rmd;
-  else
-    {
-      metadata_head = rmd;
-      metadata_tail = rmd;
-    }
+  /* raop_metadata_send() will add rtptime to these */
+  rmd->start = 0;
+  rmd->end = (duration * 44100UL) / 1000UL;
 
   return rmd;
 
@@ -1162,57 +1156,15 @@ raop_add_headers(struct raop_session *rs, struct evrtsp_request *req, enum evrts
   return 0;
 }
 
-static int
-raop_grab_cseq(struct evkeyvalq *headers)
-{
-  const char *param;
-  int cseq;
-  int ret;
-
-  param = evrtsp_find_header(headers, "CSeq");
-  if (!param)
-    return -1;
-
-  ret = safe_atoi32(param, &cseq);
-  if (ret < 0)
-    {
-      DPRINTF(E_LOG, L_RAOP, "Could not convert CSeq value to integer (%s)\n", param);
-
-      return -1;
-    }
-
-  return cseq;
-}
-
+/* This check should compare the reply CSeq with the request CSeq, but it has
+ * been removed because RAOP targets like Reflector and AirFoil don't return
+ * the CSeq according to the rtsp spec, and the CSeq is not really important
+ * anyway.
+ */
 static int
 raop_check_cseq(struct raop_session *rs, struct evrtsp_request *req)
 {
-  int reply_cseq;
-  int request_cseq;
-
-  reply_cseq = raop_grab_cseq(req->input_headers);
-  /* AirFoil won't return cseq, so skip the check */
-  if (reply_cseq < 0)
-    {
-      DPRINTF(E_INFO, L_RAOP, "No CSeq in reply, skipping check\n");
-
-      return 0;
-    }
-
-  request_cseq = raop_grab_cseq(req->output_headers);
-  if (request_cseq < 0)
-    {
-      DPRINTF(E_LOG, L_RAOP, "No CSeq in request\n");
-
-      return -1;
-    }
-
-  if (reply_cseq == request_cseq)
-    return 0;
-
-  DPRINTF(E_LOG, L_RAOP, "Reply CSeq does not match request CSeq: got %d expected %d\n", reply_cseq, request_cseq);
-
-  return -1;
+  return 0;
 }
 
 static int
@@ -2205,16 +2157,23 @@ raop_metadata_startup_send(struct raop_session *rs)
 }
 
 void
-raop_metadata_send(int id, uint64_t rtptime, uint64_t offset, int startup)
+raop_metadata_send(struct raop_metadata *rmd, uint64_t rtptime, uint64_t offset, int startup)
 {
   struct raop_session *rs;
-  struct raop_metadata *rmd;
   uint32_t delay;
   int ret;
 
-  rmd = raop_metadata_prepare(id, rtptime);
-  if (!rmd)
-    return;
+  rmd->start += rtptime;
+  rmd->end += rtptime;
+
+  /* Add the rmd to the metadata list */
+  if (metadata_tail)
+    metadata_tail->next = rmd;
+  else
+    {
+      metadata_head = rmd;
+      metadata_tail = rmd;
+    }
 
   for (rs = sessions; rs; rs = rs->next)
     {
@@ -2230,13 +2189,11 @@ raop_metadata_send(int id, uint64_t rtptime, uint64_t offset, int startup)
       if (ret < 0)
 	{
 	  raop_session_failure(rs);
-
 	  continue;
 	}
     }
 }
 
-
 /* Volume handling */
 static float
 raop_volume_convert(int volume, char *name)
@@ -2290,7 +2247,8 @@ raop_set_volume_internal(struct raop_session *rs, int volume, evrtsp_req_cb cb)
   raop_volume = raop_volume_convert(volume, rs->devname);
 
   /* Don't let locales get in the way here */
-  ret = evbuffer_add_printf(evbuf, "volume: %d.%06d\r\n", (int)raop_volume, -(int)(1000000.0 * (raop_volume - (int)raop_volume)));
+  /* We use -%d and -(int)raop_volume so -0.3 won't become 0.3 */
+  ret = evbuffer_add_printf(evbuf, "volume: -%d.%06d\r\n", -(int)raop_volume, -(int)(1000000.0 * (raop_volume - (int)raop_volume)));
   if (ret < 0)
     {
       DPRINTF(E_LOG, L_RAOP, "Out of memory for SET_PARAMETER payload (volume)\n");
